# vim: set fileencoding=utf-8 :
# Solfege - free ear training software
# Copyright (C) 2009, 2011, 2016 Tom Cato Amundsen
# License is GPL, see file COPYING


import unittest
from solfege import statistics
from solfege import application
from solfege import optionparser
from solfege.exercises import idbyname


class TestDB(unittest.TestCase):

    def test_en(self):
        s1 = "var = 1"
        s2 = "#comment\nvar = 1"
        s3 = "#comment\n\nvar = 1"
        self.assertEqual(statistics.hash_lessonfile_text(s1),
                         statistics.hash_lessonfile_text(s2))
        self.assertEqual(statistics.hash_lessonfile_text(s1),
                         statistics.hash_lessonfile_text(s3))

    def test_two(self):
        opt_parser = optionparser.SolfegeOptionParser()
        options, args = opt_parser.parse_args()
        a = application.SolfegeApp(options)
        t = idbyname.Teacher('idbyname')
        t.set_lessonfile('solfege:lesson-files/chord-min-major')
        t.m_statistics.reset_session()
        t.m_statistics.add_wrong('minor', 'major')
        t.m_statistics.add_wrong('minor', 'major')
        t.m_statistics.add_wrong('minor', 'minor')
        t.m_statistics.add_correct('major')
        for seconds in (-1, 0, 10000):
            self.assertEqual(t.m_statistics.get_num_correct_for_key(seconds, 'minor'), 1)
            self.assertEqual(t.m_statistics.get_num_guess_for_key(seconds, 'minor'), 3)

    def test_store_variables(self):
        db = statistics.DB()
        db.set_variable('database_version', 2)
        self.assertEqual(db.get_variable('database_version'), 2)
        db.set_variable('database_version', 3)
        self.assertEqual(db.get_variable('database_version'), 3)
        self.assertRaises(db.VariableTypeError,
            db.set_variable, 'database_version', '3')
        db.set_variable('textvar', 'this is кофе')
        self.assertEqual(db.get_variable('textvar'), 'this is кофе')
        # Variables should not change type
        self.assertRaises(db.VariableTypeError,
            db.set_variable, 'textvar', 5)
        # Variables should not change type
        self.assertRaises(db.VariableTypeError,
            db.set_variable, 'textvar', 5.5)
        # Variables should not change type
        self.assertRaises(db.VariableTypeError,
            db.set_variable, 'textvar', True)
        # only str is allowed, not str
        self.assertRaises(db.VariableTypeError,
            db.set_variable, 'str_text', b'bytestring')
        db.set_variable('known_number', 3.14159265)
        self.assertEqual(db.get_variable('known_number'), 3.14159265)
        self.assertRaises(db.VariableUndefinedError,
            db.get_variable, 'e')
        db.del_variable('known_number')
        self.assertRaises(db.VariableUndefinedError,
            db.del_variable, 'does_not_exist')


suite = unittest.makeSuite(TestDB)



# Solfege - free ear training software
# Copyright (C) 2004, 2007, 2008, 2011, 2016  Tom Cato Amundsen
# License is GPL, see file COPYING


class MpdException(Exception):
    """
    I think all exceptions that are raised from
    mpd.parser.parse_to_score_object should inherit from this. They will
    all be handled by abstract.Gui.standard_exception_handler.
    """
    pass



# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2007, 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Only utility functions and classes that are private to the mpd module
should go into this file.
"""

import re
from solfege.mpd.musicalpitch import MusicalPitch


def int_to_octave_notename(i):
    return MusicalPitch.new_from_int(i).get_octave_notename()


def int_to_user_octave_notename(i):
    return MusicalPitch.new_from_int(i).get_user_octave_notename()


def notename_to_int(n):
    return MusicalPitch.new_from_notename(n).semitone_pitch()


def key_to_accidentals(key):
    i = ['aeses', 'eeses', 'beses', 'fes', 'ces', 'ges', 'des', 'aes',
         'ees', 'bes', 'f', 'c', 'g', 'd', 'a', 'e', 'b', 'fis', 'cis',
         'gis', 'dis', 'ais', 'eis', 'bis'].index(key[0]) - 11
    if key[1] == 'minor':
        i = i - 3
    if i > 0:
        r = ['fis', 'cis', 'gis', 'dis', 'ais', 'eis',
             'bis', 'fis', 'cis', 'gis', 'dis'][:i]
        m = 'is'
    elif i < 0:
        r = ['bes', 'ees', 'aes', 'des', 'ges', 'ces',
             'fes', 'bes', 'ees', 'aes', 'des'][:-i]
        m = 'es'
    else:
        r = []
    retval = []
    for a in r:
        if a not in retval:
            retval.append(a)
        else:
            del retval[retval.index(a)]
            retval.append(a + m)
    return retval


def find_possible_first_note(music):
    """
    Return a tuple of 2 integer locating what we believe is the first
    pitch (but do not include the duration).

    Return the location of the text we don't understand if we are not able
    to parse the music.
    """
    i = 0
    # FIXME regexes are modified copies from the mpd Lexer. Try to reuse
    # code in the future.
    re_white = re.compile(r"\s+")
    re_clef = re.compile(r"\\clef\s+(\w*)", re.UNICODE)
    re_clef_quoted = re.compile(r"\\clef\s+\"([A-Za-z1-9]+[_^1-9]*)\"", re.UNICODE)
    re_time = re.compile(r"\\time\s+(\d+)\s*/\s*(\d+)", re.UNICODE)
    re_times = re.compile(r"\\times\s+(\d+)\s*/\s*(\d+)\s*{", re.UNICODE)
    re_key = re.compile(r"\\key\s+([a-z]+)\s*\\(major|minor)", re.UNICODE)
    re_note = re.compile("(?P<beamstart>(\[\s*)?)(?P<chordstart>(\<\s*)?)(?P<pitchname>[a-zA-Z]+[',]*)(\d+\.*)?")
    i = 0
    re_list = re_white, re_clef_quoted, re_clef, re_key, re_times, re_time, re_note
    while 1:
        for r in re_list:
            m = r.match(music[i:])
            if m:
                if r != re_note:
                    i += m.end()
                    break
                elif r == re_note:
                    assert m
                    i += len(m.group("beamstart"))
                    i += len(m.group("chordstart"))
                    return i, i + len(m.group('pitchname'))
            elif r == re_list[-1]:
                return i, i + 1



# vim: set fileencoding=utf-8:
# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2007, 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys

from gi.repository import Gtk

from solfege import cfg
from solfege import gu
from solfege import lessonfilegui


class AbstractQuestionNameTable(Gtk.Table, cfg.ConfigUtils):
    """
    Base class for QuestionNameButtonTable and QuestionNameCheckButtonTable.
    """

    def __init__(self, exname):
        Gtk.Table.__init__(self)
        cfg.ConfigUtils.__init__(self, exname)
        self._ignore_watch = 0
        self.add_watch('ask_for_names', self.ask_for_names_changed)

    def ask_for_names_changed(self):
        pass

    def initialize(self, num, dir):
        self.m_num = num
        self.m_dir = dir
        self.m_x = 0
        self.m_y = 0
        for X in self.get_children():
            X.destroy()
        self.m_button_dict = {}
        self.m_name_list = []

    def conditional_newline(self):
        """
        Do newline if there are enough buttons on the line.
        """
        if self.m_dir == 'vertic':
            self.m_y = self.m_y + 1
            if self.m_y == self.m_num:
                self.m_y = 0
                self.m_x = self.m_x + 1
        else:
            self.m_x = self.m_x + 1
            if self.m_x == self.m_num:
                self.m_x = 0
                self.m_y = self.m_y + 1

    def newline(self):
        if self.m_dir == 'vertic':
            self.m_y = 0
            self.m_x = self.m_x + 1
        else:
            self.m_x = 0
            self.m_y = self.m_y + 1

    def grab_focus_first_button(self):
        v = self.get_children()
        v.reverse()
        for c in v:
            if c.get_property("sensitive"):
                c.grab_focus()
                return


class QuestionNameButtonTable(AbstractQuestionNameTable):

    def __init__(self, exname):
        AbstractQuestionNameTable.__init__(self, exname)

    def ask_for_names_changed(self, *v):
        """
        This method is called when the config variable 'ask_for_names' is
        changed. The watching of the method is set up in
        AbstractQuestionNameTable.__init__
        """
        if self._ignore_watch > 0:
            return
        for n, button in list(self.m_button_dict.items()):
            button.set_sensitive(
                self.m_name_list.index(n) in self.get_list('ask_for_names'))

    def add(self, question, callback):
        """add a button and set up callback function.
        there should not be created more than one button with the same
        (c locale) name.
        return the button created.
        """
        if 'newline' in question and question.newline:
            self.newline()
        b = Gtk.Button()
        if question.name.cval in self.m_button_dict:
            print("Warning: The lessonfile contain several questions with the same name:", question.name.cval, file=sys.stderr)
            print("         This is a bug in the lesson file.", file=sys.stderr)
        self.m_button_dict[question.name.cval] = b
        self.m_name_list.append(question.name.cval)
        b.m_cname = question.name.cval
        b.set_sensitive(question.active)
        b.add(lessonfilegui.new_labelobject(question.name))
        b.show_all()
        self.attach(b, self.m_x, self.m_x + 1, self.m_y, self.m_y + 1)
        b.connect('clicked', callback)
        b.connect('button_release_event', callback)
        self.conditional_newline()
        return b


class QuestionNameCheckButtonTable(AbstractQuestionNameTable):

    def __init__(self, teacher):
        AbstractQuestionNameTable.__init__(self, teacher.m_exname)
        self.m_t = teacher

    def ask_for_names_changed(self, *v):
        self.m_t.m_P.m_random.reset()
        if not self.m_name_list:
            # If m_name_list is empty, it means that no lesson file is
            # yet loaded. Just return.
            return
        if self._ignore_watch > 0:
            return
        for question in self.m_t.m_P.m_questions:
            question.active = self.m_name_list.index(question.name.cval) in self.get_list('ask_for_names')

    def add(self, question):
        """add a button and set up callback function.
        there should not be created more than one button with the same
        (c locale) name.
        return the button created.
        """
        if 'newline' in question and question.newline:
            self.newline()
        b = Gtk.CheckButton()
        if question.name.cval in self.m_button_dict:
            print("Warning: The lessonfile contain several questions with the same name:", question.name.cval, file=sys.stderr)
            print("         Things will not work as normal after this.", file=sys.stderr)
        self.m_button_dict[question.name.cval] = b
        self.m_name_list.append(question.name.cval)
        b.set_active(question.active)
        b.connect('toggled', self.on_checkbutton_toggled)
        b.m_cname = question.name.cval
        b.add(lessonfilegui.new_labelobject(question.name))
        b.show_all()
        self.attach(b, self.m_x, self.m_x + 1, self.m_y, self.m_y + 1)
        self.conditional_newline()
        return b

    def on_checkbutton_toggled(self, button):
        """
        Set the content of the 'ask_for_names' config variable based on
        the active status of the check buttons.
        """
        v = []
        for i in range(len(self.m_name_list)):
            if self.m_button_dict[self.m_name_list[i]].get_active():
                v.append(i)
        self.set_list('ask_for_names', v)

    def select_all(self):
        for button in list(self.m_button_dict.values()):
            button.set_active(True)


class RandomTransposeDialog(Gtk.Dialog):

    def __init__(self, initial_value, parent):
        Gtk.Dialog.__init__(self, _("Set transposition"), parent, 0,
           (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
            Gtk.STOCK_OK, Gtk.ResponseType.OK,
            ))
        self.set_default_size(100, 100)
        dlg_vbox = gu.hig_dlg_vbox()
        self.vbox.pack_start(dlg_vbox, True, True, 0)
        xbox, vbox = gu.hig_category_vbox(_("Select how to do random transposition"))
        dlg_vbox.pack_start(xbox, True, True, 0)
        label = Gtk.Label(label=_("""You can read about the different types of transposition in the lesson file documentation available on the Help menu."""))
        spin_sg = Gtk.SizeGroup(Gtk.SizeGroupMode.HORIZONTAL)
        label.set_line_wrap(True)
        vbox.pack_start(label, True, True, 0)
        sizegroup = Gtk.SizeGroup(Gtk.SizeGroupMode.HORIZONTAL)
        self.m_buttons = {}
        self.m_spins = {}
        self.m_buttons['no'] = b1 = Gtk.RadioButton.new_with_label(None, _("No"))
        self.m_spins['no'] = []
        vbox.pack_start(b1, True, True, 0)
        self.m_buttons['yes'] = b2 = Gtk.RadioButton.new_with_label_from_widget(b1, _("Yes"))
        self.m_spins['yes'] = []
        hbox = gu.bHBox(vbox)
        label_b = Gtk.Label()
        label_b.set_alignment(0.4, 0.5)
        label_b.set_markup('<span size="xx-large">♭</span>')
        label_x = Gtk.Label()
        label_x.set_alignment(0.4, 0.5)
        label_x.set_markup('<span size="xx-large">♯</span>')
        spin_sg.add_widget(label_b)
        spin_sg.add_widget(label_x)
        hbox.pack_start(b2, True, True, 0)
        hbox.pack_start(label_b, True, True, 0)
        hbox.pack_start(label_x, True, True, 0)

        self.m_buttons['accidentals'] = Gtk.RadioButton.new_with_label_from_widget(
            b1, _("Fifths relative to C major/a minor"))
        hbox = gu.bHBox(vbox)
        hbox.pack_start(self.m_buttons['accidentals'], True, True, 0)
        self.m_spins['accidentals'] = [
            Gtk.SpinButton(adjustment=Gtk.Adjustment(0, -7, 7, 1, 1)),  # 0.0, 0),
            Gtk.SpinButton(adjustment=Gtk.Adjustment(0, -7, 7, 1, 1))]  # , 0.0, 0)]
        spin_sg.add_widget(self.m_spins['accidentals'][0])
        spin_sg.add_widget(self.m_spins['accidentals'][1])
        hbox.pack_start(self.m_spins['accidentals'][0], True, True, 0)
        hbox.pack_start(self.m_spins['accidentals'][1], True, True, 0)
        gu.SpinButtonRangeController(self.m_spins['accidentals'][0],
                                     self.m_spins['accidentals'][1],
                                     -7, 7)

        self.m_buttons['key'] = Gtk.RadioButton.new_with_label_from_widget(
            b1, _i("Fifths relative to current key"))
        hbox = gu.bHBox(vbox)
        hbox.pack_start(self.m_buttons['key'], True, True, 0)
        self.m_spins['key'] = [
            Gtk.SpinButton(adjustment=Gtk.Adjustment(0, -10, 10, 1, 1)),  # 0.0, 0),
            Gtk.SpinButton(adjustment=Gtk.Adjustment(0, -10, 10, 1, 1))]  # , 0.0, 0)]
        hbox.pack_start(self.m_spins['key'][0], True, True, 0)
        hbox.pack_start(self.m_spins['key'][1], True, True, 0)
        gu.SpinButtonRangeController(self.m_spins['key'][0],
                                     self.m_spins['key'][1],
                                     -10, 10)

        self.m_buttons['atonal'] = Gtk.RadioButton.new_with_label_from_widget(
            b1, _i("Transpose notes without changing key"))
        hbox = gu.bHBox(vbox)
        hbox.pack_start(self.m_buttons['atonal'], True, True, 0)
        self.m_spins['atonal'] = [
            Gtk.SpinButton(adjustment=Gtk.Adjustment(0, -10, 10, 1, 1)),  # 0.0, 0),
            Gtk.SpinButton(adjustment=Gtk.Adjustment(0, -10, 10, 1, 1))]  # , 0.0, 0)]
        hbox.pack_start(self.m_spins['atonal'][0], True, True, 0)
        hbox.pack_start(self.m_spins['atonal'][1], True, True, 0)
        gu.SpinButtonRangeController(self.m_spins['atonal'][0],
                                     self.m_spins['atonal'][1],
                                     -10, 10)

        self.m_buttons['semitones'] = Gtk.RadioButton.new_with_label_from_widget(
            b1, _("Semitones"))
        hbox = gu.bHBox(vbox)
        hbox.pack_start(self.m_buttons['semitones'], True, True, 0)
        self.m_spins['semitones'] = [
            Gtk.SpinButton(adjustment=Gtk.Adjustment(0, -100, 100, 1, 1)),  # , 0.0, 0),
            Gtk.SpinButton(adjustment=Gtk.Adjustment(0, -100, 100, 1, 1))]  # , 0.0, 0)]
        hbox.pack_start(self.m_spins['semitones'][0], True, True, 0)
        hbox.pack_start(self.m_spins['semitones'][1], True, True, 0)
        gu.SpinButtonRangeController(self.m_spins['semitones'][0],
                                     self.m_spins['semitones'][1],
                                     -100, 100)
        for n, w in list(self.m_buttons.items()):
            sizegroup.add_widget(w)
            for ww in self.m_spins[n]:
                ww.set_sensitive(False)
            w.connect('clicked', self.on_spins_clicked, n)
        if initial_value[0] == False:
            k = 'no'
        elif initial_value[0] == True:
            k = 'yes'
        else:
            k = initial_value[0]
        self.m_buttons[k].set_active(True)
        if k in ('accidentals', 'key', 'semitones', 'atonal'):
            t, v1, v2 = initial_value
            # FIXME Because of the RangeController, we have to do this
            # twice to be sure both values are set properly.
            self.m_spins[t][1].set_value(v2)
            self.m_spins[t][0].set_value(v1)
            self.m_spins[t][1].set_value(v2)
            self.m_spins[t][0].set_value(v1)
        self.show_all()

    def on_spins_clicked(self, w, n):
        for w in self.m_spins[n]:
            w.set_sensitive(self.m_buttons[n].get_active())

    def get_value(self):
        for n, btn in list(self.m_buttons.items()):
            if btn.get_active():
                s = n
        if s == 'yes':
            return (True, )
        elif s == 'no':
            return (False, )
        return [s, self.m_spins[s][0].get_value_as_int(), self.m_spins[s][1].get_value_as_int()]



#!/usr/bin/python
# GNU Solfege - free ear training software
# Copyright (C) 2004, 2005, 2006, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
TODO/BUGS
=========

The program only runs from sourcedir. It cannot be installed.

If you open a file that includes other files, the questions from all
files will be edited as a single file, and when you save, the include
statement will be missing, and all questions are saved in one file.
The included files are un-touched.

"""

print "The lessonfile editor is broken now."
# We use this variable because debian/rules will replace it when
# building the debian package.
solfege_version = buildinfo.VERSION_STRING

import os
import sys
prefix = os.path.split(os.path.dirname(os.path.abspath(sys.argv[0])))[0]

if sys.argv[0] == './lessonfile_editor.py':
        datadir = "."
else:
    datadir = os.path.join("@prefix@", "share", "solfege")
    sys.path.append(datadir)

import src.i18n
src.i18n.setup(prefix)

import src.lessonfile_editor_main
src.lessonfile_editor_main.main(datadir)



#!/usr/bin/python

"""
literal ::=  stringliteral | integer | floatnumber
expression_list ::=  expression ( "," expression )* [","]
assignment_stmt ::=  target "=" expression_list
target          ::=  identifier
atom      ::=  identifier | literal
# We use this for imported modules lookup
attributeref ::=  primary "." identifier
"""


class ParseTreeException(Exception):

    def __init__(self, msg, token):
        Exception.__init__(self, msg)
        self.m_tokenpos = token.m_tokenpos


class LookupException(ParseTreeException):
    pass


class Identifier(str):
    check_ns = True

    def __repr__(self):
        return "Identifier(%s)" % self

    def evaluate(self, local_namespace, global_namespace):
        if "." in self:
            mod, name = self.split(".")
        else:
            mod = None
            name = self
        for namespace in local_namespace, global_namespace:
            if mod:
                if mod in namespace:
                    return namespace[mod].m_globals[name]
            else:
                if name in namespace:
                    return namespace[name]
        if self.check_ns:
            raise LookupException("Unknown identifier '%s'" % name, self)
        return self


class Node(object):

    def dump(self, indent=0):
        print(" " * indent, self)


class Literal(Node):
    """
    Literal python object: integer or unicode string
    """

    def __init__(self, value):
        Node.__init__(self)
        assert isinstance(value, (int, str, float))
        self.m_value = value

    def evaluate(self, local_namespace, global_namespace):
        return self.m_value

    def __repr__(self):
        return "Literal(%s)" % repr(self.m_value)


class TempoType(Node):

    def __init__(self, bpm, nlen):
        self.m_bpm = bpm
        self.m_nlen = nlen

    def evaluate(self, local_namespace, global_namespace):
        return (self.m_bpm.evaluate(local_namespace, global_namespace),
               self.m_nlen.evaluate(local_namespace, global_namespace))


class Assignment(Node):

    def __init__(self, name, expression):
        Node.__init__(self)
        self.left = name
        self.right = expression

    def __repr__(self):
        return "%s = %s" % (self.left, self.right)


class FunctionCall(Node):

    class WrongArgumentCount(ParseTreeException):

        def __init__(self, functioncall):
            ParseTreeException.__init__(self, "Wrong argument count", functioncall)

    def __init__(self, dp, name, args):
        # We need to remember which dataparser called us because
        # at least one of the functions need this. FIXME I think we
        # want to remove this.
        Node.__init__(self)
        self.m_dataparser = dp
        self.m_name = name
        self.m_args = args

    def __repr__(self):
        return "%s(%s)" % (self.m_name, self.m_args)

    def evaluate(self, local_namespace, global_namespace):
        args = [x.evaluate(local_namespace, global_namespace) for x in self.m_args]
        # Functions are only in the global namespace, as functions
        # cannot be defined in the lesson files
        if self.m_name not in global_namespace:
            raise LookupException("Function name unknown", self)
        if global_namespace[self.m_name][0]:
            args.insert(0, self.m_dataparser)
        try:
            return global_namespace[self.m_name][1](*args)
        except TypeError:
            raise self.WrongArgumentCount(self.m_args)


class Addition(Node):

    def __init__(self, left, right):
        self.left = left
        self.right = right

    def evaluate(self, local_namespace, global_namespace):
        return (self.left.evaluate(local_namespace, global_namespace)
                + self.right.evaluate(local_namespace, global_namespace))


class StringFormatting(Node):

    def __init__(self, left, right):
        self.left = left
        self.right = right

    def evaluate(self, local_namespace, global_namespace):
        return (self.left.evaluate(local_namespace, global_namespace)
                % self.right.evaluate(local_namespace, global_namespace))


class ExpressionList(Node, list):

    def evaluate(self, local_namespace, global_namespace):
        return [x.evaluate(local_namespace, global_namespace) for x in self]


class CodeBlock(Node, list):

    def add_statement(self, statement):
        self.append(statement)

    def dump(self, indent=0):
        Node.dump(self, indent)
        try:
            for statement in self:
                statement.dump(indent + 2)
        except AttributeError:
            pass


class IncludeStatement(Node):

    def __init__(self, inctree):
        Node.__init__(self)
        self.m_inctree = inctree


class Block(CodeBlock):

    def __init__(self, blocktype):
        CodeBlock.__init__(self)
        self.m_blocktype = blocktype

    def __repr__(self):
        return "UnnamedBlock(type=%s)" % self.m_blocktype


class NamedBlock(CodeBlock):
    """
    A named block will have the name referring to it inserted
    into the namespace it is defined in.
    """

    def __init__(self, blocktype, name):
        CodeBlock.__init__(self)
        self.m_blocktype = blocktype
        self.m_name = name

    def __repr__(self):
        return "NamedBlock(type=%s, name=%s)" % (self.m_blocktype, self.m_name)


class Program(CodeBlock):

    def __init__(self):
        CodeBlock.__init__(self)

    def dump(self, indent=0):
        print("Program:")
        for statement in self:
            statement.dump(indent + 2)



# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2007, 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import operator

import cairo

from solfege.mpd import const
from solfege.mpd import elems
from solfege.mpd import mpdutils
from solfege.mpd.musicalpitch import MusicalPitch
from functools import reduce

ACCIDENTAL__2 = -2
ACCIDENTAL__1 = -1
ACCIDENTAL_0 = 0
ACCIDENTAL_1 = 1
ACCIDENTAL_2 = 2

NOTEHEAD_WHOLE = 0
NOTEHEAD_HALF = 1
NOTEHEAD_QUARTER = 2

FLAG_8UP = 'u3'
FLAG_8DOWN = 'd3'
FLAG_16UP = 'u4'
FLAG_16DOWN = 'd4'
FLAG_32UP = 'u5'
FLAG_32DOWN = 'd5'
FLAG_64UP = 'u6'
FLAG_64DOWN = 'd6'

fetadir = "feta"


class dim20:
    linespacing = 8
    stemlen = linespacing * 3.5
    xshift = 10
    col_width = 20
    first_staff_ypos = 80
    staff_spacing = 100
    ledger_left = -4
    ledger_right = 14
    accidental_widths = {-2: 12, -1: 7, 0: 5, 1: 8, 2: 8}
    clef_yoffset = {
        'G': -38 + 3 * linespacing,
        'F': -8 + 3 * linespacing,
        'C': -15 + 3 * linespacing,
    }

dimentions = {20: dim20}

accidental_y_offset = {ACCIDENTAL__2: -7,
                       ACCIDENTAL__1: -7,
                       ACCIDENTAL_0: -5,
                       ACCIDENTAL_1: -6,
                       ACCIDENTAL_2: -2}


class Engraver:

    def __init__(self):
        self.m_fontsize = 20

    def get_width(self):
        """
        Get the width of all engravers that does not implement their
        own get_width function.
        """
        return 20

    def __str__(self):
        return '(%s)' % self.__class__


class ClefEngraver(Engraver):

    def __init__(self, clef):
        Engraver.__init__(self)
        self.m_clef = clef

    def get_width(self):
        return 25

    def engrave(self, ct, staff_yoffset):
        ims = cairo.ImageSurface.create_from_png(
            fetadir + '/feta%i-clefs-%s.png'
            % (self.m_fontsize, self.m_clef.get_symbol()))
        ct.set_source_surface(
            ims, self.m_xpos,
            dimentions[self.m_fontsize].clef_yoffset[self.m_clef.get_symbol()]
            - self.m_clef.get_stafflinepos() * dimentions[self.m_fontsize].linespacing
            + staff_yoffset)
        ct.paint()


class TimeSignatureEngraver(Engraver):

    def __init__(self, timesig):
        Engraver.__init__(self)
        self.m_timesig = timesig

    def engrave(self, ct, staff_yoffset):
        x = 0
        for idx in range(len(str(self.m_timesig.m_num))):
            n = str(self.m_timesig.m_num)[idx]
            ims = cairo.ImageSurface.create_from_png(
                fetadir + '/feta%i-number-%s.png' % (self.m_fontsize, n))
            ct.set_source_surface(ims, self.m_xpos + x, staff_yoffset - 12)
            x += 10
            ct.paint()
        x = 0
        for idx in range(len(str(self.m_timesig.m_den))):
            n = str(self.m_timesig.m_den)[idx]
            ims = cairo.ImageSurface.create_from_png(
                fetadir + '/feta%i-number-%s.png' % (self.m_fontsize, n))
            ct.set_source_surface(ims, self.m_xpos + x, staff_yoffset + 3)
            x += 10
            ct.paint()

    def __str__(self):
        return '(TimeSignatureEngraver:%i/%i)' % (self.m_timesig.m_den,
            self.m_timesig.m_num)


class TieEngraver(Engraver):

    def __init__(self, note1, note2):
        Engraver.__init__(self)
        self.m_note1 = note1
        self.m_note2 = note2

    def engrave(self, ct, staff_yoffset):
        dim = dimentions[self.m_fontsize]
        p1 = self.m_note1.m_xpos + dim.xshift
        p2 = self.m_note2.m_xpos
        w = p2 - p1 - dim.xshift
        w = max(w, dim.xshift)
        ct.arc((p1 + (p2 - p1) / 2),
               int(staff_yoffset + dim.linespacing * (self.m_note2.m_ypos - 1) / 2 + 3),
               int(w), 30 * 3.14 / 180, 150 * 3.14 / 180)
        ct.stroke()


class AccidentalsEngraver(Engraver):

    def __init__(self, accs):
        Engraver.__init__(self)
        self.m_accs = accs

        def f(v, w=dimentions[self.m_fontsize].accidental_widths):
            x = 0
            for a in v:
                x += w[a]
            return x
        self.m_usize = reduce(operator.__add__, list(map(f, list(accs.values())))) + len(accs)

    def get_width(self):
        return self.m_usize

    def engrave(self, ct, staff_yoffset):
        x = 0
        for y in self.m_accs:
            for acc in self.m_accs[y]:
                ims = cairo.ImageSurface.create_from_png(
                        fetadir + '/feta%i-accidentals-%i.png' %
                            (self.m_fontsize, acc))
                ct.set_source_surface(ims, self.m_xpos + x,
                        int(accidental_y_offset[acc] + staff_yoffset
                         + dimentions[self.m_fontsize].linespacing * y / 2
                         + accidental_y_offset[acc]))
                x += dimentions[self.m_fontsize].accidental_widths[acc] + 1
        ct.paint()


class KeySignatureEngraver(Engraver):

    def __init__(self, old_key, key, clef):
        Engraver.__init__(self)
        self.m_clef = clef
        self.m_old_accidentals = mpdutils.key_to_accidentals(old_key)
        self.m_accidentals = mpdutils.key_to_accidentals(key)

    def get_width(self):
        # FIXME this value depends on the fontsize, and on what kind of
        # accidental we are drawing. A sharp is wider that a flat.
        return len(self.m_accidentals) * 10 + len(self.m_old_accidentals) * 6 + 8

    def engrave(self, ct, staff_yoffset):
        x = 0
        dolist = []
        # natural signs
        for acc in self.m_old_accidentals:
            if acc in self.m_accidentals:
                continue
            ypos = self.m_clef.an_to_ylinepos(acc)
            dolist.append((0, x, ypos))
            # FIXME see FIXME msg in .get_width
            x += 6
        # accidentals
        for acc in self.m_accidentals:
            if acc.endswith('eses'):
                ktype = -2
            elif acc.endswith('es'):
                ktype = -1
            elif acc.endswith('isis'):
                ktype = 2
            else:
                ktype = 1
            ypos = self.m_clef.an_to_ylinepos(acc)
            dolist.append((ktype, x, ypos))
            # FIXME see FIXME msg in .get_width
            x += 10
        for ktype, x, ypos in dolist:
            ims = cairo.ImageSurface.create_from_png(
                fetadir + '/feta%i-accidentals-%i.png' % (self.m_fontsize, ktype))
            ct.set_source_surface(ims, self.m_xpos + x,
                int(accidental_y_offset[ktype] + staff_yoffset
                     + dimentions[self.m_fontsize].linespacing * ypos / 2
                     + accidental_y_offset[ktype]))
            ct.paint()


class NoteheadEngraver(Engraver):

    def __init__(self, shift, ypos, musicalpitch, duration):
        """
        m_ypos == 0 is the middle line on the staff.
        Negative value is up, positive is down.
        The value counts notesteps, not pixels!
        FIXME, right now it also draws dots, but they should be an own class
        because the dots has to be handled special with noteheads are xshifted.
        """
        Engraver.__init__(self)
        if duration.m_nh < 2:
            self.m_head = NOTEHEAD_WHOLE
        elif duration.m_nh > 2:
            self.m_head = NOTEHEAD_QUARTER
        else:
            self.m_head = NOTEHEAD_HALF
        self.m_shift = shift
        self.m_numdots = duration.m_dots
        self.m_ypos = ypos
        self.m_midi_int = int(musicalpitch)

    def engrave(self, ct, staff_yoffset):
        dim = dimentions[self.m_fontsize]
        # Have to adjust wholenotes a little to the left to be on the middle
        # of the ledger line.
        if self.m_head == 0:
            xx = -2
        else:
            xx = 0
        ims = cairo.ImageSurface.create_from_png(
            fetadir + '/feta%i-noteheads-%i.png' % (self.m_fontsize, self.m_head))
        ct.set_source_surface(ims, self.m_xpos + self.m_shift * dim.xshift + xx,
            int(staff_yoffset + dim.linespacing * self.m_ypos / 2 - 4))
        ct.paint()
        for n in range(self.m_numdots):
            ims = cairo.ImageSurface.create_from_png(
                fetadir + '/feta20-dots-dot.png')
            ct.set_source_surface(ims, int(self.m_xpos + dim.xshift * (self.m_shift + 1.5 + n / 2.0)),
                -3 + staff_yoffset + dim.linespacing * self.m_ypos / 2)
        ct.paint()

    def __str__(self):
        try:
            xpos = self.m_xpos
        except AttributeError:
            xpos = None
        return "(NoteheadEngraver: xpos:%s, ypos:%i, head:%i, shift=%i)" % (
                xpos, self.m_ypos, self.m_head, self.m_shift)

    def __repr__(self):
        try:
            xpos = self.m_xpos
        except AttributeError:
            xpos = None
        return "<NoteheadEngraver xpos:%s, ypos:%i, head:%i, shift=%i>" % (
                xpos, self.m_ypos, self.m_head, self.m_shift)


class BarlineEngraver(Engraver):

    def __init__(self, barline_type):
        Engraver.__init__(self)
        self.m_type = barline_type

    def get_width(self):
        return 8

    def engrave(self, ct, staff_yoffset):
        dim = dimentions[self.m_fontsize]
        ct.move_to(self.m_xpos + 0.5, staff_yoffset - dim.linespacing * 2)
        ct.line_to(self.m_xpos + 0.5, staff_yoffset + dim.linespacing * 2)


class TupletEngraver(Engraver):

    def __init__(self, ratio, direction):
        Engraver.__init__(self)
        self.m_stems = []
        self.m_ratio = ratio
        self.m_direction = direction

    def add_stem(self, stem):
        self.m_stems.append(stem)

    def do_layout(self):
        dim = dimentions[self.m_fontsize]
        top = []
        bottom = []
        for s in self.m_stems:
            if s.m_stemdir == const.UP:
                top.append(min(s.m_yposes) - 7)
                bottom.append(max(s.m_yposes) + 1)
            else:
                top.append(min(s.m_yposes))
                bottom.append(max(s.m_yposes) + 5)
        self.m_t = min(top) - 2
        self.m_b = max(bottom) + 2
        self.m_xpos1 = self.m_stems[0].m_xpos - 2
        self.m_xpos2 = self.m_stems[-1].m_xpos + 2
        if self.m_stems[-1].m_stemdir == const.DOWN:
            self.m_xpos2 += dim.xshift

    def engrave(self, ct, staff_yoffset):
        self.do_layout()
        dim = dimentions[self.m_fontsize]
        m = (self.m_xpos1 + self.m_xpos2) / 2
        if self.m_direction == const.UP or self.m_direction == const.BOTH:
            y = min(staff_yoffset - dim.linespacing * 3,
                   staff_yoffset + self.m_t * dim.linespacing / 2)
            d = 1
        else:  # == const.DOWN
            y = max(staff_yoffset + dim.linespacing * 5,
                    staff_yoffset + self.m_b * dim.linespacing / 2)
            d = -1
        self.m_xpos1 += 0.5
        self.m_xpos2 += 0.5
        if len(self.m_stems) > 1:
            # left horiz
            ct.move_to(self.m_xpos1, y)
            ct.line_to(m - 6, y)
            # right horiz
            ct.move_to(m + 6, y)
            ct.line_to(self.m_xpos2, y)
            # left vertic
            ct.move_to(self.m_xpos1, y)
            ct.line_to(self.m_xpos1, y + 5 * d)
            # right vertic
            ct.move_to(self.m_xpos2, y)
            ct.line_to(self.m_xpos2, y + 5 * d)
            ct.stroke()
        ct.select_font_face("Sans 10")
        tt = ct.text_extents(str(self.m_ratio.m_den))
        ct.move_to(m - tt[2] / 2, y + tt[3] / 2)
        ct.show_text(str(self.m_ratio.m_den))
        ct.stroke()


class BeamEngraver(Engraver):

    def __init__(self):
        Engraver.__init__(self)
        self.m_stems = []

    def add_stem(self, stem_engraver):
        self.m_stems.append(stem_engraver)

    def do_layout(self):
        # We don't beam single notes. Notes can be single when the dictation
        # exercise tries to display the first note
        if len(self.m_stems) == 1:
            self.m_stems[0].m_is_beamed = False
        self.decide_beam_stemdir()
        self.set_stemlens(self.find_lowhigh_ypos())

    def set_stemlens(self, lh):
        l, h = lh
        for e in self.m_stems:
            if self.m_stemdir == const.UP:
                e.m_beamed_stem_top = l - 6
            else:
                assert self.m_stemdir == const.DOWN
                e.m_beamed_stem_top = h + 6

    def find_lowhigh_ypos(self):
        """
        Find the lowest and highest notehead in the beam.
        """
        mn = 1000
        mx = -1000
        for se in self.m_stems:
            for ylinepos in se.m_yposes:
                if mn > ylinepos:
                    mn = ylinepos
                if mx < ylinepos:
                    mx = ylinepos
        return mn, mx

    def decide_beam_stemdir(self):
        """
        Decide the direction for the stems in this beam, and set
        the stemdir for all stems.
        """
        v = {const.UP: 0, const.DOWN: 0}
        for e in self.m_stems:
            v[e.m_stemdir] = v[e.m_stemdir] + 1
            e.m_stemlen = 50
        if v[const.UP] > v[const.DOWN]:
            stemdir = const.UP
        else:
            stemdir = const.DOWN
        for e in self.m_stems:
            e.m_stemdir = stemdir
        self.m_stemdir = stemdir

    def engrave(self, ct, staff_yoffset):
        # Do nothing if there are only one stem in the beam. Usually this
        # does not happen, but it is possible that a lesson file want to
        # display only the first note in some music, and that note is beamed.
        if len(self.m_stems) == 1:
            return
        dim = dimentions[self.m_fontsize]
        d = 0
        if self.m_stemdir == const.UP:
            d = 1
        else:
            d = -1
        x1 = self.m_stems[0].m_xpos
        x2 = self.m_stems[-1].m_xpos
        if self.m_stems[0].m_beamed_stem_top % 2 == 1:
            for x in range(len(self.m_stems)):
                self.m_stems[x].m_beamed_stem_top = \
                    self.m_stems[x].m_beamed_stem_top - d
        y1 = self.m_stems[0].m_beamed_stem_top * dim.linespacing // 2 + staff_yoffset
        beamw = 3
        beaml = 10
        for y in range(beamw):
            ct.move_to(x1, y1 + y * d)
            ct.line_to(x2, y1 + y * d)
        for stem in self.m_stems:
            stem._beam_done = 8

        def short_beam(stem, xdir, beamnum, d=d, beamw=beamw, beaml=beaml, y1=y1):
            for y in range(beamw):
                ct.move_to(stem.m_xpos, y1 + y + beamnum * d * beamw * 2)
                ct.line_to(stem.m_xpos + xdir * beaml, y1 + y + beamnum * d * beamw * 2)
        for nl, yc in ((16, 0), (32, 1), (64, 2)):
            for i in range(len(self.m_stems) - 1):
                if self.m_stems[i].m_duration.m_nh >= nl \
                        and self.m_stems[i + 1].m_duration.m_nh >= nl:
                    for y in range(beamw):
                        ct.move_to(self.m_stems[i].m_xpos, d * beamw * yc * 2 + y1 + y + d * beamw * 2)
                        ct.line_to(self.m_stems[i + 1].m_xpos, d * beamw * yc * 2 + y1 + y + d * beamw * 2)
                    self.m_stems[i]._beam_done = nl
                    self.m_stems[i + 1]._beam_done = nl
                if self.m_stems[i].m_duration.m_nh >= nl \
                      and self.m_stems[i + 1].m_duration.m_nh <= nl / 2 \
                      and self.m_stems[i]._beam_done < nl:
                    if i == 0:
                        stemdir = 1
                    else:
                        if self.m_stems[i - 1].m_duration.m_nh < \
                           self.m_stems[i + 1].m_duration.m_nh:
                            stemdir = 1
                        else:
                            stemdir = -1
                    short_beam(self.m_stems[i], stemdir, yc + 1, d)
                    self.m_stems[i]._beam_done = nl
            if self.m_stems[-1].m_duration.m_nh >= nl \
                    and self.m_stems[-1]._beam_done <= nl / 2:
                short_beam(self.m_stems[-1], -1, yc + 1, d)
        ct.stroke()


class StemEngraver(Engraver):
    """
    Every notehead belong to a stem, even if the stem is invisible.
    """

    def __init__(self, noteheads, duration, stemdir, is_beamed):
        """
        If stemdir == BOTH: Look at the location of the noteheads, and descide
        which way the stem will go. Set m_stemdir to UP or DOWN.
        """
        Engraver.__init__(self)
        self.m_noteheads = noteheads
        # FIXME maybe remove, if m_noteheads will be available anyway
        self.m_yposes = sorted([h.m_ypos for h in noteheads])
        self.m_duration = duration
        self.m_stemdir = stemdir
        self.m_is_beamed = is_beamed
        # descide the stem direction
        if len(self.m_yposes) == 1:
            x = self.m_yposes[0]
        else:
            x = self.m_yposes[0] + self.m_yposes[-1]
        if self.m_stemdir == const.BOTH:
            if x >= 0:
                self.m_stemdir = const.UP
            else:
                self.m_stemdir = const.DOWN
        self.shift_noteheads()
        self.m_stemlen = dimentions[self.m_fontsize].stemlen

    def get_width(self):
        c = 2
        if [n for n in self.m_noteheads if n.m_shift == -1]:
            c += 1
        if [n for n in self.m_noteheads if n.m_shift == 1]:
            c += 1
        return c * dimentions[self.m_fontsize].xshift

    def calc_xpos(self):
        """
        TODO: if we readd m_stempos, it has to be taken into consideration here.
        """
        if self.m_stemdir == const.UP:
            self.m_xpos += dimentions[self.m_fontsize].xshift
        if [n for n in self.m_noteheads if n.m_shift == -1]:
            self.m_xpos += dimentions[self.m_fontsize].xshift
            for n in self.m_noteheads:
                n.m_xpos += dimentions[self.m_fontsize].xshift

    def engrave(self, ct, staff_yoffset):
        dim = dimentions[self.m_fontsize]
        if self.m_duration.m_nh < 2:
            return
        # draw flags
        # we must save the context state because set_source_surface
        # makes the line we draw for stems invisible
        ct.save()
        if not self.m_is_beamed and self.m_duration.m_nh > 4:
            if self.m_stemdir == const.UP:
                ims = cairo.ImageSurface.create_from_png(
                    fetadir + '/feta%i-flags-%s.png' %
                    (self.m_fontsize, self.get_flag(const.UP)))
                ct.set_source_surface(ims, self.m_xpos,
                    int(staff_yoffset - self.m_stemlen + self.m_yposes[0] * dim.linespacing / 2))
            else:
                ims = cairo.ImageSurface.create_from_png(
                    fetadir + '/feta%i-flags-%s.png'
                    % (self.m_fontsize, self.get_flag(const.DOWN)))
                ct.set_source_surface(ims, self.m_xpos,
                    int(staff_yoffset + 4 + self.m_yposes[-1] * dim.linespacing / 2))
            ct.paint()
        ct.restore()
        # draw stem
        if self.m_stemdir == const.DOWN:
            if self.m_is_beamed:
                yroot = self.m_beamed_stem_top
            else:
                yroot = self.m_yposes[-1] + 6
            ytop = self.m_yposes[0]
        else:
            if self.m_is_beamed:
                yroot = self.m_beamed_stem_top
            else:
                yroot = self.m_yposes[0] - 6
            ytop = self.m_yposes[-1]
        ct.move_to(self.m_xpos + 0.5, ytop * dim.linespacing / 2 + staff_yoffset)
        ct.line_to(self.m_xpos + 0.5, yroot * dim.linespacing / 2 + staff_yoffset)
        ct.stroke()

    def get_flag(self, stemdir):
        assert self.m_duration.m_nh > 4
        return {8: {const.UP: FLAG_8UP,
                   const.DOWN: FLAG_8DOWN},
                16: {const.UP: FLAG_16UP,
                     const.DOWN: FLAG_16DOWN},
                32: {const.UP: FLAG_32UP,
                     const.DOWN: FLAG_32DOWN},
                64: {const.UP: FLAG_64UP,
                     const.DOWN: FLAG_64DOWN}} \
                              [self.m_duration.m_nh][stemdir]

    def shift_noteheads(self):
        """
        We need to look at all stems at the same timepos in the staff when
        shifting note heads. An integer value say if and in which direction
        a note head is shifted. The value 0 means that the note head is in
        its default position. In default position, the noteheads
        on stems pointing UP are to the left of the stem and noteheads on
        stems pointing DOWN on the right of the stem. And they are all
        aligned below each other.

        A positive shift value means that the head is moved to the right,
        and a negative is to the left. The unit is notehead-widths, so
        the common (and for now only used) value is -1 for noteheads on
        down-stems and +1 for noteheads on up-stems.
        """
        self.m_noteheads.sort(key=lambda n: n.m_ypos)
        # TODO: handle crashing noteheads in two different voices.

        # When laying out noteheads, we start with the notehead at the
        # end of the stem. That will always be unshifted.
        if self.m_stemdir == const.UP:
            heads = list(reversed(self.m_noteheads))
        else:
            heads = self.m_noteheads
        last_ypos = None
        for idx, head in enumerate(heads):
            if last_ypos is not None:
                if abs(last_ypos - head.m_ypos) == 1 and heads[idx - 1].m_shift == 0:
                    head.m_shift = 1 if self.m_stemdir == const.UP else -1
                elif last_ypos == head.m_ypos:
                    head.m_shift = 1 if self.m_stemdir == const.UP else -1
            last_ypos = head.m_ypos

    def __str__(self):
        return "(StemEngraver)"

    def __repr__(self):
        return "<StemEngraver xpos:%s>" % (
                self.m_xpos, )


class LedgerLineEngraver(Engraver):

    def __init__(self, up, down):
        """
        up: number of ledger lines above staff
        down: number of ledger lines below staff
        """
        Engraver.__init__(self)
        self.m_xpos = 0
        self.m_up = up
        self.m_down = down

    def engrave(self, ct, staff_yoffset):
        dim = dimentions[self.m_fontsize]
        if self.m_up:
            for y in range(self.m_up):  # FIXME use rel_line_to
                ct.move_to(self.m_xpos + dim.ledger_left, (-y - 3) * dim.linespacing + staff_yoffset)
                ct.line_to(self.m_xpos + dim.ledger_right, (-y - 3) * dim.linespacing + staff_yoffset)
        if self.m_down:
            for y in range(self.m_down):
                ct.move_to(self.m_xpos + dim.ledger_left, (y + 3) * dim.linespacing + staff_yoffset)
                ct.line_to(self.m_xpos + dim.ledger_right, (y + 3) * dim.linespacing + staff_yoffset)
        ct.stroke()

    def __str__(self):
        return "(LedgerLineEngraver xpos:%i, updown%i%i" % (
            self.m_xpos, self.m_up, self.m_down)

    def __repr__(self):
        return "<LedgerLineEngraver xpos:%i, updown%i:%i>" % (
            self.m_xpos, self.m_up, self.m_down)


class RestEngraver(Engraver):

    def __init__(self, ypos, dur):
        Engraver.__init__(self)
        self.m_ypos = ypos
        self.m_dots = dur.m_dots
        self.m_type = {1: 0, 2: 1, 4: 2, 8: 3, 16: 4, 32: 5, 64: 6}[dur.m_nh]

    def get_width(self):
        # FIXME write me!
        return 20

    def engrave(self, ct, staff_yoffset):
        dim = dimentions[self.m_fontsize]
        if self.m_type == 0:
            my = dim.linespacing / 2
        else:
            my = 0
        ims = cairo.ImageSurface.create_from_png(
            fetadir + '/feta%i-rests-%i.png' % (self.m_fontsize, self.m_type))
        ct.set_source_surface(
            ims, self.m_xpos,
            int(staff_yoffset - my + dim.linespacing * self.m_ypos / 2 - 4))
        ct.paint()
        for n in range(self.m_dots):
            ims = cairo.ImageSurface.create_from_png(fetadir + '/feta20-dots-dot.xpm')
            ct.set_source_surface(
                int(self.m_xpos + dim.xshift * (1.5 + n / 2.0)),
                -3 + staff_yoffset + dim.linespacing * self.m_ypos / 2)
            ct.paint()

    def __str__(self):
        return "(RestEngraver)"


class SkipEngraver(Engraver):

    def __init__(self, duration):
        Engraver.__init__(self)
        self.m_duration = duration

    def engrave(self, ct, staff_yoffset):
        pass


class _StaffCommon(list):

    def __init__(self, staff, last_timepos):
        """
        Create engraver objects for the staff, one timepos at the time.
        Append them to self, and add refs to them in the m_engravers dict
        to access them by timepos and type.
        """
        list.__init__(self)
        self.m_label = getattr(staff, 'm_label', None)
        self.m_engravers = {}
        # When joining two empty Scores with Score.concat2, we need to return
        # to avoid IndexErrors further down this method.
        if not staff.w_parent().m_bars:
            return
        # make a set of all timeposes in the staffs voices
        t = set()
        # We need to add the timepos of the beginning of all bars, since
        # all staffs has to display the time signature if it changes. Normally
        # it is not necessary to do this here, since staff.m_tdict will have
        # the timepos. But it is necessary for Scores created by Score.concat2
        [t.add(b.m_timepos) for b in staff.w_parent().m_bars]
        # Then we add the timepos of all notes and rests
        [t.add(tp) for tp in staff.m_tdict]
        for voice in staff.m_voices:
            [t.add(tp) for tp in voice.m_tdict]
        if last_timepos is None:
            # display all notes
            timeposes = sorted(t)
        else:
            timeposes = [x for x in sorted(t) if x < last_timepos]
        #
        clef = None
        keysig = ("c", "major")
        self.refill_accidentals_info(keysig)
        for voice in staff.m_voices:
            # tmp variable needed to keep track of the beams. Deleted at method exit.
            voice.m_beam = None
            voice.m_ties = {}
            # tuplet
            voice.m_tuplet = None
        bar_idx = 0
        props = {
            'hide-timesignature': False,
        }
        beams = []
        for timepos in timeposes:
            if (bar_idx < len(staff.w_parent().m_bars) - 1
                  and timepos == staff.w_parent().m_bars[bar_idx + 1].m_timepos):
                bar_idx += 1
            if timepos not in self.m_engravers:
                self.m_engravers[timepos] = {}
            eng = self.m_engravers[timepos]
            ##############
            # Properties #
            ##############
            if (timepos in staff.m_tdict
                    and 'properties' in staff.m_tdict[timepos]):
                props.update(staff.m_tdict[timepos]['properties'])
            # Forget accidentals at bar lines
            if timepos == staff.w_parent().m_bars[bar_idx].m_timepos:
                self.refill_accidentals_info(keysig)
            ########
            # Clef #
            ########
            if timepos in staff.m_tdict and 'clef' in staff.m_tdict[timepos]:
                clef = staff.m_tdict[timepos]['clef']
                eng['clef'] = ClefEngraver(staff.m_tdict[timepos]['clef'])
                self.append(eng['clef'])
            #################
            # Key signature #
            #################
            if timepos in staff.m_tdict and 'keysig' in staff.m_tdict[timepos]:
                eng['keysig'] = KeySignatureEngraver(keysig,
                    staff.m_tdict[timepos]['keysig'], clef)
                self.append(eng['keysig'])
                keysig = staff.m_tdict[timepos]['keysig']
                self.refill_accidentals_info(keysig)
            ##################
            # Time signature #
            ##################
            if not staff.w_parent().m_bars:
                return
            if (props['hide-timesignature'] == False
                and ((timepos == staff.w_parent().m_bars[bar_idx].m_timepos
                and bar_idx > 0
                and staff.w_parent().m_bars[bar_idx].m_timesig
                 != staff.w_parent().m_bars[bar_idx - 1].m_timesig) or timepos == elems.TimeSignature(0, 1))):
                eng['timesig'] = TimeSignatureEngraver(staff.w_parent().m_bars[bar_idx].m_timesig)
                self.append(eng['timesig'])
            ###############
            # Accidentals #
            ###############
            if isinstance(self, StaffContext):
                v = {}
                for voice in staff.m_voices:
                    if timepos not in voice.m_tdict:
                        continue
                    # If the 'elem' is a rest or skip,
                    # then there are no accidentals.
                    if isinstance(voice.m_tdict[timepos]['elem'][0],
                            (elems.Rest, elems.Skip)):
                        continue
                    for elem in voice.m_tdict[timepos]['elem']:
                        e = self.needed_accidental(elem.m_musicalpitch)
                        if e is not None:
                            v[clef.steps_to_ylinepos(elem.m_musicalpitch.steps())] = e
                if v:
                    self.append(AccidentalsEngraver(v))
                    eng['accidentals'] = self[-1]
            ############################################
            # Create stems, noteheads and ledger lines #
            ############################################
            # These two count show many ledger lines we need.
            yline_up = 0
            yline_down = 0
            for voice in staff.m_voices:
                if timepos not in voice.m_tdict:
                    continue
                if 'elem' in voice.m_tdict[timepos]:
                    elem = voice.m_tdict[timepos]['elem']
                    if isinstance(elem, elems.Stem):
                        if elem.m_beaminfo == 'start':
                            voice.m_beam = BeamEngraver()
                            beams.append(voice.m_beam)
                            self.append(voice.m_beam)
                        # If the tuplet contain only one tone, then elem.m_tupletinfo == 'end' and
                        # voice.m_tuplet will be None
                        if elem.m_tupletinfo == 'start' or (elem.m_tupletinfo == 'end' and voice.m_tuplet is None):
                            voice.m_tuplet = TupletEngraver(elem.m_tuplet_ratio, elem.m_tuplet_dir)
                    if 'elem' not in eng:
                        eng['elem'] = []
                    if isinstance(voice.m_tdict[timepos]['elem'][0], elems.Rest):
                        e = RestEngraver(0, voice.m_tdict[timepos]['elem'][0].m_duration)
                        self.append(e)
                        eng['elem'].append(e)
                    elif isinstance(elem[0], elems.Skip):
                        e = SkipEngraver(elem[0].m_duration)
                        self.append(e)
                        eng['elem'].append(e)
                    elif not isinstance(voice.m_tdict[timepos]['elem'][0], elems.Skip):
                        elist, stemengraver = self.create_notehead_engraver(clef, voice.m_tdict[timepos]['elem'])
                        for note, engraver in zip(voice.m_tdict[timepos]['elem'], elist):
                            if note.m_tieinfo == 'start':
                                voice.m_ties[note.m_musicalpitch.get_octave_notename()] = engraver
                            elif note.m_tieinfo == 'go':
                                self.append(TieEngraver(voice.m_ties[note.m_musicalpitch.get_octave_notename()], engraver))
                                del voice.m_ties[note.m_musicalpitch.get_octave_notename()]
                                voice.m_ties[note.m_musicalpitch.get_octave_notename()] = engraver
                            elif note.m_tieinfo == 'end':
                                self.append(TieEngraver(voice.m_ties[note.m_musicalpitch.get_octave_notename()], engraver))
                                del voice.m_ties[note.m_musicalpitch.get_octave_notename()]
                        if voice.m_beam:
                            stemengraver.m_is_beamed = True
                            voice.m_beam.add_stem(stemengraver)
                        if voice.m_tuplet:
                            voice.m_tuplet.add_stem(stemengraver)
                        eng['elem'].extend(elist)
                        eng['elem'].append(stemengraver)
                        self.extend(elist)
                        self.append(stemengraver)
                    if isinstance(elem, elems.Stem):
                        if elem.m_beaminfo == 'end':
                            voice.m_beam = None
                        if elem.m_tupletinfo == 'end':
                            self.append(voice.m_tuplet)
                            voice.m_tuplet = None
                # Ledger lines
                for elem in voice.m_tdict[timepos]['elem']:
                    if isinstance(elem, elems.Note):
                        if not clef:
                            # clef is None on a rhythm staff. Then we need no
                            # ledger lines
                            continue
                        ypos = clef.steps_to_ylinepos(elem.m_musicalpitch.steps())
                        if yline_up > ypos < -5:
                            yline_up = ypos
                        if yline_down < ypos > 5:
                            yline_down = ypos
            if yline_up:
                yline_up = - yline_up // 2 - 2
            if yline_down:
                yline_down = yline_down // 2 - 2
            if yline_up or yline_down:
                e = LedgerLineEngraver(yline_up, yline_down)
                eng['elem'].append(e)
                self.append(e)
        # We do this here instead of further up where we check for'
        # if m_beaminfo == 'end' because we need to do do_layout for beams
        # even when we only want to engrave the first note in a beam.
        for b in beams:
            b.do_layout()


class StaffContext(_StaffCommon):

    def refill_accidentals_info(self, key):
        """Fill the .m_accidentals_info dict with the accidentals
        that exist in the key signature `key`.
        """
        self.m_accidentals_info = {}
        for step in range(MusicalPitch.LOWEST_STEPS, MusicalPitch.HIGHEST_STEPS + 1):
            self.m_accidentals_info[step] = 0
        for a in mpdutils.key_to_accidentals(key):
            n = MusicalPitch.new_from_notename(a)
            for octave in range(-4, 7):
                n.m_octave_i = octave
                if n.semitone_pitch() < 128:
                    if a[-4:] == 'eses':
                        self.m_accidentals_info[n.steps()] = -2
                    elif a[-2:] == 'es':
                        self.m_accidentals_info[n.steps()] = -1
                    elif a[-4:] == 'isis':
                        self.m_accidentals_info[n.steps()] = 2
                    else:
                        self.m_accidentals_info[n.steps()] = 1

    def needed_accidental(self, m):
        steps = m.steps()
        if m.m_accidental_i != self.m_accidentals_info[steps]:
            if (self.m_accidentals_info[steps] == 2 and m.m_accidental_i == 1) \
                    or (self.m_accidentals_info[steps] == -2 and m.m_accidental_i == -1):
                self.m_accidentals_info[steps] = m.m_accidental_i
                return [0, m.m_accidental_i]
            self.m_accidentals_info[steps] = m.m_accidental_i
            return [m.m_accidental_i]

    def create_notehead_engraver(self, clef, elements):
        """
        Return a two item
        1: Return a list of the engravers that draw the note heads
        2: The stem engraver
        Will return the engravers in the same order as the notes in elements.
        """
        noteheads = []
        for elem in elements:
            noteheads.append(NoteheadEngraver(
            0,  # shift,
            clef.steps_to_ylinepos(elem.m_musicalpitch.steps()),  # ypos,
            elem.m_musicalpitch,
            elem.m_duration))
        return (noteheads,
            StemEngraver(noteheads, elements[0].m_duration, elements.m_stemdir, False))


class RhythmStaffContext(_StaffCommon):
    """
    Typeset notes on a single line. Pitches are ignored.
    """

    def refill_accidentals_info(self, key):
        """
        nop since we don't care about pitches.
        """
        return

    def create_notehead_engraver(self, clef, elem):
        """
        Return a two item
        1: Return a list of the engravers that draw the note heads
        2: The stem engraver
        """
        notehead = NoteheadEngraver(
            0,  # shift,
            0,  # ypos,
            elem[0].m_musicalpitch,
            elem[0].m_duration)
        return [notehead], StemEngraver([notehead], elem[0].m_duration, elem.m_stemdir, False)


class ScoreContext(object):

    def __init__(self, score, last_timepos=None):
        t = set()
        self.m_contexts = staff_contexts = []
        # Create one staff context for each staff line.
        for staff in score.m_staffs:
            if isinstance(staff, elems.RhythmStaff):
                staff_contexts.append(RhythmStaffContext(staff, last_timepos))
            else:
                staff_contexts.append(StaffContext(staff, last_timepos))
            [t.add(timepos) for timepos in staff_contexts[-1].m_engravers]
        # BarlineEngravers have the timepos of the beginning of the next
        # bar. So the last bar line will have the timepos of where the
        # bar after the last bar would begin. We need to add that timepos
        # to the set 't' because it is not added by the above loop.
        if score.m_bars:
            t.add(score.m_bars[-1].end())
        # Initialize the property dict for each staff.
        for staff_context in staff_contexts:
            staff_context.props = {
                'hide-barline': False,
            }
        xpos = 0
        bar_idx = 0
        # This loop will set the m_xpos variable for all engraver objects
        for timepos in sorted(t):
            ########################
            # Per score properties #
            ########################
            for staff_context, staff in zip(staff_contexts, score.m_staffs):
                if (timepos in staff.m_tdict
                        and 'properties' in staff.m_tdict[timepos]):
                    staff_context.props.update(staff.m_tdict[timepos]['properties'])
            if timepos == score.m_bars[bar_idx].end():
                for staff_context in staff_contexts:
                    if not staff_context.props['hide-barline']:
                        e = BarlineEngraver("|")
                        staff_context.append(e)
                        staff_context.m_engravers.setdefault(
                            score.m_bars[bar_idx].end(), {})['barline'] = e
                bar_idx += 1

            def do_col(s, xpos):
                max_width = 0
                for context in staff_contexts:
                    if timepos not in context.m_engravers:
                        continue
                    if s in context.m_engravers[timepos]:
                        context.m_engravers[timepos][s].m_xpos = xpos
                        max_width = max(max_width, context.m_engravers[timepos][s].get_width())
                return max_width
            xpos += do_col('barline', xpos)
            xpos += do_col('clef', xpos)
            xpos += do_col('keysig', xpos)
            xpos += do_col('timesig', xpos)
            xpos += do_col('accidentals', xpos)
            # Elems
            max_width = 0
            for context in staff_contexts:
                if timepos not in context.m_engravers:
                    continue
                if 'elem' in context.m_engravers[timepos]:
                    for e in context.m_engravers[timepos]['elem']:
                        e.m_xpos = xpos
                        if isinstance(e, StemEngraver):
                            e.calc_xpos()
                    max_width = max(max_width, max(
                        [e.get_width() for e in context.m_engravers[timepos]['elem']]))
            xpos += max_width



# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2006, 2007, 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
FIXME update.
In addition to the names in this file, only this is public in
the mpd module:

  Classes:
    MusicalPitch
    MusicDisplayer
"""


LOWEST_NOTENAME = "c,,,,"
HIGHEST_NOTENAME = "g''''''"

from solfege.mpd.duration import Duration
from solfege.mpd.musicalpitch import MusicalPitch, InvalidNotenameException
from solfege.mpd.interval import Interval
from solfege.mpd.mpdutils import notename_to_int, int_to_octave_notename, int_to_user_octave_notename
from solfege.mpd.parser import parse_to_score_object
from solfege.mpd.rat import Rat
from solfege.mpd.track import Track, PercussionTrack
from solfege.mpd._exceptions import MpdException
from solfege.mpd.performer import score_to_tracks
try:
    # solfege.app_running is set in solfege/startup.py, but will be
    # unset if run from some of the build scripts. This way we can
    # build without having a X display.
    import solfege
    if solfege.app_running:
        from solfege.mpd.rhythmwidget import RhythmWidget, RhythmWidgetController
        from solfege.mpd.musicdisplayer import MusicDisplayer
except AttributeError:
    pass
finally:
    del solfege


def music_to_tracklist(music, start=None, end=None):
    """
    return a list of tracks, where track[0] use only channel 0,
    track[1] only use channel 1 etc.
    """
    return score_to_tracks(parse_to_score_object(music), start, end)


def music_to_track(music, start=None, end=None):
    tracklist = score_to_tracks(parse_to_score_object(music), start, end)
    track = tracklist[0]
    for x in range(1, len(tracklist)):
        track.merge_with(tracklist[x])
    return track


##################
# midi functions #
##################

def transpose_notename(n, t):
    assert isinstance(n, str)
    assert isinstance(t, int)
    # 1 2 sekund
    # 3 4 ters
    # 5 6 kvart
    # 7   kvint
    # 8 9 sekst
    # 10 11 septim
    return int_to_octave_notename(notename_to_int(n) + t)


def compare_notenames(n1, n2):
    return notename_to_int(n1) - notename_to_int(n2)


def select_clef(s):
    """
    argument s is a string with notenames like this: " c e g c' f' g''"
    """
    lowest = HIGHEST_NOTENAME
    highest = LOWEST_NOTENAME
    for n in s.split():
        if compare_notenames(n, lowest) < 0:
            lowest = n
        if compare_notenames(n, highest) > 1:
            highest = n
    if compare_notenames(highest, "c'") < 0:
        return "bass"
    if compare_notenames(lowest, "c'") >= 0:
        return "violin"
    if compare_notenames(highest, "c'") >= compare_notenames("c'", lowest):
        return "violin"
    else:
        return "bass"



# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2007, 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


def gcd(a, b):
    while b:
        a, b = b, a % b
    return a


class Rat(object):
    __slots__ = ('m_num', 'm_den')
    """
    This is NOT a generic rational number class, it includes
    only the features needed for the the soundcard and mpd
    module included in Solfege.
    """

    def __init__(self, num, den=1):
        assert isinstance(num, int)
        assert isinstance(den, int)
        self.m_num = num
        self.m_den = den

    def clone(self):
        return Rat(self.m_num, self.m_den)

    def __repr__(self):
        return "(Rat %i/%i)" % (self.m_num, self.m_den)

    def __str__(self):
        return "(Rat %i/%i)" % (self.m_num, self.m_den)

    def __add__(self, B):
        assert isinstance(self, Rat)
        assert isinstance(B, Rat)
        a = self.m_num * B.m_den + B.m_num * self.m_den
        b = self.m_den * B.m_den
        g = gcd(a, b)
        return Rat(a // g, b // g)

    def __sub__(self, B):
        assert isinstance(self, Rat)
        assert isinstance(B, Rat)
        a = (self.m_num * B.m_den - B.m_num * self.m_den)
        b = self.m_den * B.m_den
        g = gcd(a, b)
        return Rat(a // g, b // g)

    def __mul__(self, B):
        assert isinstance(self, Rat)
        if isinstance(B, int):
            g = gcd(self.m_num * B, self.m_den)
            return Rat(self.m_num * B // g, self.m_den // g)
        assert isinstance(B, Rat)
        g = gcd(self.m_num * B.m_num, self.m_den * B.m_den)
        return Rat(self.m_num * B.m_num // g, self.m_den * B.m_den // g)

    def __truediv__(self, i):
        if isinstance(i, int):
            i = Rat(i, 1)
        r = Rat(self.m_num * i.m_den, self.m_den * i.m_num)
        g = gcd(r.m_num, r.m_den)
        return Rat(r.m_num // g, r.m_den // g)

    def __rmul__(self, B):
        """ called when integer * Rat
        """
        assert isinstance(self, Rat)
        assert isinstance(B, int)
        a = self.m_num * B
        g = gcd(a, self.m_den)
        return Rat(a // g, self.m_den // g)

    def __int__(self):
        return self.m_num // self.m_den

    def __float__(self):
        return self.m_num / self.m_den

    def __hash__(self):
        return hash((self.m_num, self.m_den))

    def __eq__(self, B):
        if B is None:
            return False
        return self.m_num / self.m_den == B.m_num / B.m_den

    def __lt__(self, B):
        return self.m_num / self.m_den < B.m_num / B.m_den

    def __le__(self, B):
        return self.m_num / self.m_den <= B.m_num / B.m_den



# Solfege - free ear training software
# Copyright (C) 2007, 2008, 2011, 2016 Tom Cato Amundsen
# License is GPL, see file COPYING


import unittest


class TestSourceCode(unittest.TestCase):
    pass
suite = unittest.makeSuite(TestSourceCode)



# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2006, 2007, 2008, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


r"""
REMEMBER: down is positive, up is negative.

All voices begin at the beginning of the staff. It is
not possible to split a voice in two like in Lilypond.

The parser will not handle fis and f in the same octave on one stem.

Rules:
 * It can be different timesignatures in different staffs.
 * \key has to come before \time

The parser does not care if you have correct number of notes in a bar.
To get bar lines you have to insert a '|'
"""

import logging

from solfege.mpd.duration import Duration
from solfege.mpd.lexer import Lexer
from solfege.mpd.musicalpitch import MusicalPitch, InvalidNotenameException
from solfege.mpd.rat import Rat
from solfege.mpd import const
from solfege.mpd import _exceptions
from solfege.mpd import elems


class ParseError(_exceptions.MpdException):

    def __init__(self, msg, lexer):
        _exceptions.MpdException.__init__(self, msg)
        self.m_lineno, self.m_linepos1, self.m_linepos2 = lexer.get_error_location()


def musicalpitch_relative(first, second):
    """
    think:  \relative c'{ first second }
    Tritone handling is the same as GNU Lilypond

    I placed here instead of in MusicalPitch since it is only used
    once in parse_to_score_object and I don't think anyone need this
    in MusicalPitch.
    """
    assert isinstance(first, MusicalPitch)
    assert isinstance(second, MusicalPitch)
    n1 = second.clone()
    n1.m_octave_i = first.m_octave_i
    n2 = n1.clone()
    if n1 < first:
        n2.m_octave_i += 1
    else:
        n1.m_octave_i -= 1
    if n2.steps() - first.steps() < first.steps() - n1.steps():
        # we go  up
        n2.m_octave_i += second.m_octave_i
        return n2
    else:
        # we go down
        n1.m_octave_i += second.m_octave_i
        return n1


class TimeSignature:

    def __init__(self, num, den):
        self.m_num = num
        self.m_den = den


def parse_to_score_object(music):
    lexer = Lexer(music)
    relative_mode = None
    relto = None
    transpose_pitch = None
    TOPLEVEL = 1  # 'toplevel'
    NOTES = 2  # 'notes'
    START_OF_CHORD = 3  # 'start-of-chord'
    CHORD = 4  # 'chord'
    context = TOPLEVEL
    score = elems.Score()
    chord_duration = None
    cur_duration = Duration(4, 0)
    tie_is_in_the_air = 0
    beam = None
    # None when not parsing notes in a tuplet. Item 0 is the ration and 1.. is the notes
    times = None
    cur_staff = None
    # we set it just so pyflakes does not complain.
    relto_backup = None
    # This variable is set to the duration of the pickup bar from we parse
    # \partial nn until the bar has been created.
    partial = None
    for toc, toc_data in lexer:
        try:
            if toc_data.m_duration:
                cur_duration = toc_data.m_duration.clone()
        except AttributeError:
            pass
        if toc == Lexer.STAFF:
            assert context == TOPLEVEL
            cur_staff = score.add_staff(elems.Staff)
            cur_voice = cur_staff.m_voices[-1]
            stem_dir = const.BOTH
            tuplet_dir = const.BOTH
            relative_mode = None
            timepos = Rat(0)
            last_pos = timepos
        elif toc == Lexer.RHYTHMSTAFF:
            assert context == TOPLEVEL
            cur_staff = score.add_staff(elems.RhythmStaff)
            cur_voice = cur_staff.m_voices[-1]
            stem_dir = const.BOTH
            tuplet_dir = const.BOTH
            relative_mode = None
            timepos = Rat(0)
            last_pos = timepos
        elif toc == Lexer.VOICE:
            if not cur_staff:
                raise ParseError("Don't use \\addvoice before \\staff", lexer)
            relative_mode = None
            timepos = Rat(0)
            cur_voice = cur_staff.add_voice()
        elif toc == Lexer.RELATIVE:
            assert not relative_mode
            relative_mode = 1
            relto = toc_data
        elif toc == Lexer.TRANSPOSE:
            transpose_pitch = toc_data
        elif toc == Lexer.PARTIAL:
            partial = toc_data
        elif toc == Lexer.TIME:
            if not cur_staff:
                raise ParseError("\\time can not be used before \\staff", lexer)
            # FIXME
            # Also now, we only allow the first voice to change time signatures
            if cur_staff.m_voices.index(cur_voice) != 0:
                raise ParseError("only timesig in first voice", lexer)
            # FIXME: we are stricter with time signatures that both solfege 3.16
            # and LilyPond
            if not cur_voice.is_bar_full():
                raise ParseError("timesig change only when bar is full!", lexer)
            if partial:
                score.add_partial_bar(partial, toc_data)
                partial = None
            else:
                score.add_bar(toc_data)
        elif toc == Lexer.KEY:
            p = MusicalPitch.new_from_notename(toc_data[0])
            if transpose_pitch:
                p.transpose_by_musicalpitch(transpose_pitch)
            k = (p.get_notename(), toc_data[1])
            if not cur_staff:
                raise ParseError("\\key can not be used before \\staff", lexer)
            cur_staff.set_key_signature(k, timepos)
        elif toc == Lexer.TIMES:
            if not times:
                times = [toc_data]
            else:
                raise ParseError(r"\times nn/nn does not nest", lexer)
        elif toc == Lexer.CLEF:
            try:
                cur_staff.set_clef(toc_data, timepos)
            except elems.UnknownClefException as e:
                e.m_lineno, e.m_linepos1, e.m_linepos2 = lexer.get_error_location()
                raise
        elif toc == '|':
            if timepos != score.get_bar_at(last_pos).end():
                logging.warning("Bar check failed at %s", timepos)
        elif toc == '{':
            if (context == TOPLEVEL):
                context = NOTES
                # if not cur_staff.m_coldict[Rat(0, 1)].m_keysignature:
                # FIXME dont understand
                if transpose_pitch:
                    k = (transpose_pitch.get_notename(), 'major')
                else:
                    k = ('c', 'major')
                cur_voice.set_key_signature(k)
            else:
                raise ParseError("Token '{' not allowed here.", lexer)
        elif toc == '<':
            if context == NOTES:
                context = START_OF_CHORD
            else:
                raise ParseError("Token '<' not allowed here.", lexer)
        elif toc == '>':
            if context == CHORD:
                if tie_is_in_the_air:
                    # The 3.16-parser only handles ties between whole chords, not
                    # single tones of a chord.
                    if tie_is_in_the_air:
                        for last_note in cur_voice.m_tdict[last_pos]['elem']:
                            for cur_note in cur_voice.m_tdict[timepos]['elem']:
                                if last_note.m_musicalpitch.get_octave_notename() == cur_note.m_musicalpitch.get_octave_notename():
                                    cur_voice.tie([last_note, cur_note])
                    tie_is_in_the_air = 0
                last_pos = timepos
                timepos = timepos + chord_duration.get_rat_value()
                chord_duration = None
                relto = relto_backup
                relto_backup = None
                context = NOTES
            else:
                raise ParseError("Token '>' not allowed here.", lexer)
        elif toc == '}':
            if context == NOTES:
                if times:
                    cur_voice.tuplet(times[0], tuplet_dir, times[1:])
                    times = None
                    cur_duration.m_tuplet = Rat(1, 1)
                else:
                    context = TOPLEVEL
            else:
                raise ParseError("Token '}' not allowed here.", lexer)
        elif toc == '[':
            beam = []
        elif toc == ']':
            cur_voice.beam(beam)
            beam = None
        elif toc == '~':
            tie_is_in_the_air = 1
        elif toc == Lexer.NOTE and (context in [NOTES, CHORD, START_OF_CHORD]):
            # FIXME check if toc_data.m_duration will ever be undefined.
            # If not we can do this:   if not toc_data.m_duration:
            if not getattr(toc_data, 'm_duration', None):
                toc_data.m_duration = cur_duration.clone()
            if times:
                toc_data.m_duration.m_tuplet = times[0].clone()
            if relative_mode:
                toc_data.m_pitch = musicalpitch_relative(
                                          relto, toc_data.m_pitch)
                relto = toc_data.m_pitch.clone()
            if transpose_pitch:
                toc_data.transpose(transpose_pitch)
            if partial:
                score.add_partial_bar(partial, None)
                partial = None
            if context == NOTES:
                note = elems.Note(toc_data.m_pitch, toc_data.m_duration)
                try:
                    cur_voice.append(note, stem_dir)
                except elems.Voice.BarFullException as e:
                    raise ParseError(str(e), lexer)
                if beam is not None:
                    beam.append(note)
                if times is not None:
                    times.append(note)
                # The 3.16-parser only handles ties between whole chords, not
                # single tones of a chord.
                if tie_is_in_the_air:
                    for note in cur_voice.m_tdict[last_pos]['elem']:
                        for n in cur_voice.m_tdict[timepos]['elem']:
                            if n.m_musicalpitch.get_octave_notename() == note.m_musicalpitch.get_octave_notename():
                                cur_voice.tie([note, n])
                    tie_is_in_the_air = 0
                last_pos = timepos
                timepos = timepos + toc_data.m_duration.get_rat_value()
            elif context == START_OF_CHORD:
                cur_voice.append(elems.Note(toc_data.m_pitch, toc_data.m_duration), stem_dir)
                relto_backup = relto
                chord_duration = toc_data.m_duration
                context = CHORD
            elif context == CHORD:
                cur_voice.add_to(timepos, elems.Note(toc_data.m_pitch, toc_data.m_duration))
        elif toc == Lexer.SKIP and context == NOTES:
            if toc_data.m_duration:
                cur_duration = toc_data.m_duration.clone()
            else:
                toc_data.m_duration = cur_duration.clone()
            skip = elems.Skip(toc_data.m_duration)
            cur_voice.append(skip)
            last_pos = timepos
            timepos = timepos + toc_data.m_duration.get_rat_value()
        elif toc == Lexer.REST and context == NOTES:
            if toc_data.m_duration:
                cur_duration = toc_data.m_duration.clone()
            else:
                toc_data.m_duration = cur_duration.clone()
            rest = elems.Rest(toc_data.m_duration)
            cur_voice.append(rest)
            last_pos = timepos
            timepos += toc_data.m_duration.get_rat_value()
        elif toc == Lexer.STEMDIR:
            stem_dir = toc_data
        elif toc == Lexer.TUPLETDIR:
            tuplet_dir = toc_data
        else:
            raise ParseError(toc, lexer)
    return score


def validate_only_notenames(s):
    """
    Return (None, None, None) if the string s is only notenames
    (pitch and duration). No ties or other tokens are allowed.
    """
    lex = Lexer(s)
    try:
        for toc, toc_data in lex:
            if not toc:
                break
            if toc != lex.NOTE:
                return lex.get_error_location()
    except InvalidNotenameException:
        return lex.get_error_location()
    return None, None, None



# Solfege - free ear training software
# Copyright (C) 2007, 2008, 2011, 2016 Tom Cato Amundsen
# License is GPL, see file COPYING


import unittest
from solfege.frontpage import *


class TestLearningTree(unittest.TestCase):

    def test_create_page(self):
        page = Page('noname')
        self.assertEqual(len(page), 0)
        self.assertEqual(page.m_name, 'noname')
        page2 = Page('noname', [Column()])
        self.assertEqual(len(page2), 1)
        Page()

    def test_column(self):
        # empty col
        col = Column()
        linklist = LinkList('Intervals', [])
        # can also construct with children
        col = Column(linklist)
        self.assertEqual(len(col), 1)

    def test_add_linklist(self):
        page = Page('noname', [Column()])
        column = page[-1]
        ll = column.add_linklist('Intervals')
        self.assertEqual(ll.m_name, 'Intervals')
        self.assertFalse(ll)

    def test_linklist(self):
        ll = LinkList('Intervals', [])
        self.assertEqual(len(ll), 0)
        ll = LinkList('Intervals', ['sdfsf23'])
        self.assertEqual(len(ll), 1)
        self.assertEqual(ll[0], 'sdfsf23')
        self.assertEqual(ll, ['sdfsf23'])
        ll.append('47d82-f32sd93-f8sd9s32')
        ll.append(Page('noname'))
        self.assertEqual(ll.m_name, 'Intervals')
        self.assertEqual(len(ll), 3)

    def test_is_empty(self):
        p = Page()
        self.assertEqual(p.is_empty(), True)
        p.append(Column())
        self.assertEqual(p.is_empty(), True)
        p[0].append(LinkList('test', []))
        self.assertEqual(p.is_empty(), True)

    def test_load_tree(self):
        load_tree("exercises/standard/learningtree.txt")

    def test_iterate_filenames(self):
        p = Page('noname', [
            Column(
                LinkList('heading', [
                    'id1', 'id2', 'id3', ]),
            ),
        ])
        self.assertEqual(list(p.iterate_filenames()),
                ['id1', 'id2', 'id3'])
        p[0].append(LinkList('heading', ['id1', 'id5']))
        self.assertEqual(list(p.iterate_filenames()),
                ['id1', 'id2', 'id3', 'id1', 'id5'])

    def test_use_count(self):
        p = Page('noname', [
            Column(
                LinkList('heading', [
                    'id1', 'id2', 'id3', ]),
            ),
        ])
        self.assertEqual(p.get_use_dict(), {'id1': 1, 'id2': 1, 'id3': 1})
        p[0].append(LinkList('heading', ['id1', 'id5']))
        self.assertEqual(p.get_use_dict(), {'id1': 2, 'id2': 1, 'id3': 1, 'id5': 1})

    def test_iterate_topics_for_file(self):
        p = Page('noname', [
            Column(
                LinkList('heading', [
                    'id1', 'id2', 'id3', ]),
            ),
        ])
        p[0].append(LinkList('heading2', ['id1', 'id5']))
        self.assertEqual(list(p.iterate_topics_for_file('id1')),
                ['heading', 'heading2'])
        self.assertEqual(list(p.iterate_topics_for_file('id3')),
                ['heading'])
        self.assertEqual(list(p.iterate_topics_for_file('id5')),
                ['heading2'])


suite = unittest.makeSuite(TestLearningTree)



# GNU Solfege - free ear training software
# Copyright (C) 2005, 2006, 2007, 2008, 2011, 2016 Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


r"""
  linux                 win 3.8                           win 3.9
    ~                   C:\D&S\User
~/.solfegerc            C:\D&S\User\.solfegerc            %APPDATA\GNU Solfege\solfegerc
~/.solfege              C:\D&S\USer\.solfege              %APPDATA%\GNU Solfege
~/.solfege/trainingsets C:\D&S\User\.solfege\trainingsets My Documents\GNU Solfege\trainingsets
~/.solfege/eartr.tests  C:\D&S\User\.solfege\eartr.tests  My Documents\GNU Solfege\eartr.tests
~/lessonfiles           C:\D&S\User\lessonfiles           My Documents\GNU Solfege\lessonfiles

MSWIN: app_data() => %APPDATA%\GNU Solfege
       rcfile() => app_data()\solfegerc
       user_data() => MyDocuments\GNU Solfege
Linux: app_data() => ~/.solfege
       rcfile() => ~/.solfegerc
       user_data() => app_data()

Example locations:
    app_data()/learningtrees
    app_data()/testresults
    app_data()/statistics

    user_data()/trainingsets
    user_data()/eartrainingtests
"""

import locale
import os
import sys
if sys.platform == 'win32':
    from solfege import winreg
# This name will be used as folder name on MS Windows, to create one
# folder in the "My Documents" folder, and one folder in "Application Data"
appname = "GNU Solfege"


def get_home_dir():
    ''' Try to find user's home directory, otherwise return current directory.'''
    path1 = os.path.expanduser("~")
    try:
        path2 = os.environ["HOME"]
    except KeyError:
        path2 = ""
    try:
        path3 = os.environ["USERPROFILE"]
    except KeyError:
        path3 = ""

    if not os.path.exists(path1):
        if not os.path.exists(path2):
            if not os.path.exists(path3):
                return os.getcwd()
            else: return path3
        else: return path2
    else: return path1


def win32_program_files_folder():
    """
    Return the name of the C:\Program files folder.
    """
    return winreg._get_reg_value(winreg._winreg.HKEY_LOCAL_MACHINE,
        r"SOFTWARE\Microsoft\Windows\CurrentVersion", r"ProgramFilesDir")


# FIXME remove??
def expanduser(s):
    return s.replace("~", get_home_dir())


def user_data():
    """
    Return the full path name of a directory where solfege  by default will suggest to
    place the directories trainingsets/ eartrainingtests/
    """
    if sys.platform == "win32":
        return os.path.join(winreg._get_reg_user_value(winreg.SHELL_FOLDERS, 'Personal'), appname)
    else:
        return os.path.expanduser("~/.solfege")


def rcfile():
    """
    Return the full name of the users rc file.
    """
    if sys.platform == "win32":
        return os.path.join(winreg.get_appdata(), appname, "solfegerc")
    else:
        return os.path.expanduser("~/.solfegerc")


def app_data():
    """
    Return the full path name of the directory that will store files
    created by the program. For example statistics and test results.
    """
    if sys.platform == "win32":
        return os.path.join(winreg.get_appdata(), appname)
    else:
        return os.path.expanduser("~/.solfege")



# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2007, 2008, 2011, 2016  Tom Cato Amundsen
# Copyright (C) 2001 Joe Lee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
from solfege.soundcard.synth_common import SynthCommon
from solfege.soundcard import winmidi
from solfege.mpd.track import MidiEventStream


class WinSynth(SynthCommon):
    NUM_CHANNELS = 16

    def __init__(self, devnum, verbose_init):
        SynthCommon.__init__(self)
        try:
            self.__driver = winmidi.Winmidi(devnum)
        except RuntimeError as e:
            if not winmidi.output_devices():
                e.args = e.args + (_("No MIDI output devices available."),)
                raise
            devnum = 0
            logging.error("WinSynth.__init__: No MIDI output devices available on %i. Retrying with devNum 0", devnum)
            self.__driver = winmidi.Winmidi(devnum)
        self.m_type_major = "win32"  # FIXME
        self.m_devnum = devnum
        if verbose_init:
            logging.debug("Solfege will use Windows multimedia output.")

    def close(self):
        self.__driver = None

    def stop(self):
        # dummy function
        pass

    def play_track(self, *tracks):
        self.play_midieventstream(MidiEventStream(*tracks))

    def play_midieventstream(self, midieventstream):
        if self.__driver is None:
            raise RuntimeError("Attempted to use synth after closing.")
        self.__driver.reset(self.m_devnum)
        # bigger magic plays slower
        magic = 1440000
        self.__driver.set_tempo(int(magic * 4 / 60))
        notelen = 0
        for e in midieventstream:
            if e[0] == midieventstream.TEMPO:
                self.__driver.set_tempo(int(magic * e[2] / e[1]))
            elif e[0] == midieventstream.NOTELEN_TIME:
                notelen = e[1]
            elif e[0] == midieventstream.NOTE_ON:
                self.__driver.note_on(int(1000 * notelen), e[1], e[2], e[3])
                notelen = 0
            elif e[0] == midieventstream.NOTE_OFF:
                self.__driver.note_off(int(1000 * notelen), e[1], e[2], e[3])
                notelen = 0
            elif e[0] == midieventstream.SET_PATCH:
                self.__driver.program_change(e[1], e[2])
                # Give the synth a little time to process the program change
                self.__driver.note_off(100, 0, 0, 0)
            elif e[0] == midieventstream.BENDER:
                logging.debug("ugh todo: seq_bender for play_with_drvmidi")
                #m.seq_bender(DEV, e[1], e[2])
            else:
                pass
        self.__driver.play()



# vim: set fileencoding=utf-8:
# GNU Solfege - free ear training software
# Copyright (C) 2010, 2011, 2016  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from solfege import abstract
from solfege.mpd import elems
from solfege import gu
from solfege import lessonfile
from solfege import mpd
from solfege import soundcard


class Teacher(abstract.Teacher):

    def __init__(self, exname):
        abstract.Teacher.__init__(self, exname)
        self.lessonfileclass = lessonfile.QuestionsLessonfile

    def new_question(self):
        if self.get_bool('config/picky_on_new_question') \
                and self.q_status in [self.QSTATUS_NEW, self.QSTATUS_WRONG]:
            return self.ERR_PICKY
        self.q_status = self.QSTATUS_NEW
        self.m_P.select_random_question()

    def play_question(self):
        # We can only create countin for questions that are generated using
        # the mpd module
        music = self.m_P.get_question()['music']
        if isinstance(music, lessonfile.MpdParsable):
            score = self.m_P.get_score('music')
            countin = self.m_P.m_globals.get('countin')
            countin = self.m_P.get_question().get('countin', countin)
            if countin:
                score = elems.Score.concat2(countin.get_score(self.m_P, as_name='countin'), score)
            tracks = mpd.score_to_tracks(score)
            tracks[0].prepend_bpm(*self.m_P.get_tempo())
            soundcard.synth.play_track(*tracks)
        else:
            self.m_P.play_question()

    def get_empty_staff(self):
        """
        Return a score with and empty staff with the correct numbers of bars
        required to answer the current question.
        """
        score = elems.Score()
        score.add_staff(staff_class=elems.RhythmStaff)
        if 'rhythm' in self.m_P.get_question():
            music_obj = self.m_P.get_question()['rhythm']
            if not isinstance(music_obj, lessonfile.MpdParsable):
                raise lessonfile.LessonfileException(_("The music object named «rhythm» is not a subclass of MpdParsable"),
                _('Read about the «rhythm» variable in the section «Question block» in the chapter named «Extending GNU Solfege» in the user manual.'))
            music = music_obj.get_mpd_music_string(self.m_P)
        else:
            music_obj = self.m_P.get_question()['music']
            if not isinstance(music_obj, lessonfile.MpdParsable):
                raise lessonfile.LessonfileException(_("The music object named «music» is not a subclass of MpdParsable"),
                _('Read about the «rhythm» variable in the section «Question block» in the chapter named «Extending GNU Solfege» in the user manual.'))
            music = music_obj.get_mpd_music_string(self.m_P)
        for bar in mpd.parser.parse_to_score_object(music).m_bars:
            if isinstance(bar, elems.PartialBar):
                score.add_partial_bar(bar.m_duration, bar.m_timesig)
            else:
                score.add_bar(bar.m_timesig)
        score.voice11.fill_with_skips()
        return score

    def guess_answer(self, score):
        assert self.q_status not in (self.QSTATUS_NO, self.QSTATUS_GIVE_UP)
        q = self.m_P.get_question()
        if 'rhythm' in q:
            qvar = 'rhythm'
        else:
            qvar = 'music'
        qscore = mpd.parser.parse_to_score_object(
            q[qvar].get_mpd_music_string(self.m_P))
        if qscore.get_timelist() == score.get_timelist():
            self.q_status = self.QSTATUS_SOLVED
            return True
        else:
            self.q_status = self.QSTATUS_WRONG
            return False

    def give_up(self):
        self.q_status = self.QSTATUS_GIVE_UP


class Gui(abstract.LessonbasedGui):

    def __init__(self, teacher):
        abstract.LessonbasedGui.__init__(self, teacher)
        self.g_w = mpd.RhythmWidget()
        self.g_w.connect('score-updated', self.on_score_updated)
        self.practise_box.pack_start(self.g_w, False, False, 0)
        self.g_c = mpd.RhythmWidgetController(self.g_w)
        self.practise_box.pack_start(self.g_c, False, False, 0)
        self.g_flashbar = gu.FlashBar()
        self.practise_box.pack_start(self.g_flashbar, False, False, 0)
        self.g_flashbar.show()
        self.std_buttons_add(
            ('new', self.new_question),
            ('guess_answer', self.guess_answer),
            ('repeat', self.repeat_question),
            ('give_up', self.give_up))
        self.g_w.show()

    def on_score_updated(self, w):
        self.g_guess_answer.set_sensitive(bool(self.g_w.m_score.get_timelist()))

    def new_question(self, *w):
        def exception_cleanup():
            self.m_t.q_status = self.QSTATUS_NO
            soundcard.synth.stop()
            self.std_buttons_exception_cleanup()
        self.m_t.new_question()
        try:
            self.m_t.play_question()
            self.std_buttons_new_question()
            self.g_w.grab_focus()
            self.g_w.set_score(self.m_t.get_empty_staff())
            self.g_c.set_editable(True)
        except Exception as e:
            if isinstance(e, mpd.MpdException):
                if 'm_mpd_badcode' not in dir(e):
                    e.m_mpd_badcode = self.m_t.m_P.get_question()[e.m_mpd_varname].get_err_context(e, self.m_t.m_P)
            if not self.standard_exception_handler(e, exception_cleanup):
                raise

    def guess_answer(self, *w):
        if self.m_t.q_status == Teacher.QSTATUS_SOLVED:
            if self.m_t.guess_answer(self.g_w.m_score):
                self.g_flashbar.flash(_("Correct, but you have already solved this question"))
            else:
                self.g_flashbar.flash(_("Wrong, but you have already solved this question"))
        else:
            if self.m_t.guess_answer(self.g_w.m_score):
                self.g_flashbar.flash(_("Correct"))
                self.std_buttons_answer_correct()
            else:
                self.g_flashbar.flash(_("Wrong"))
                self.std_buttons_answer_wrong()
                self.g_w.grab_focus()

    def repeat_question(self, *w):
        self.g_w.grab_focus()
        self.m_t.play_question()

    def give_up(self, *w):
        if 'rhythm' in self.m_t.m_P.get_question():
            varname = 'rhythm'
        else:
            varname = 'music'
        answer = mpd.parser.parse_to_score_object(
                self.m_t.m_P.get_question()[varname].get_mpd_music_string(
                    self.m_t.m_P))
        answer.m_staffs.append(self.g_w.m_score.m_staffs[0])
        answer.m_staffs[-1].set_parent(answer)
        answer.m_staffs[0].m_label = _("The music played:")
        answer.m_staffs[-1].m_label = _("The rhythm you entered:")
        answer.create_shortcuts()
        self.g_w.set_score(answer, cursor=None)
        self.m_t.give_up()
        self.g_c.set_editable(False)
        self.std_buttons_give_up()

    def on_start_practise(self):
        super(Gui, self).on_start_practise()
        self.std_buttons_start_practise()
        self.g_c.set_editable(False)
        self.g_w.set_score(elems.Score())
        self.g_flashbar.delayed_flash(self.short_delay,
            _("Click 'New' to begin."))



#!/usr/bin/python
"""
Used to sanity check prefixes used in lesson files.
"""

import os
import sys
import re
import glob
import optparse

parser = optparse.OptionParser()
parser.add_option("-l", action="store_true", dest="list_strings",
    help="List the strings we find.")
parser.add_option("-v", action="store_true", dest="list_prefixes",
    help="List all prefixes")
parser.add_option("-p", dest="string_prefix",
    help="The 'PREFIX|xxx' to seach for.")
parser.add_option("-f", action="store_true", dest="display_filename",
    help="Show the name of the files the PREFIX is found in.")

(options, args) = parser.parse_args()

r = re.compile("\"%s\|(?P<prefix>.*?)\"" % options.string_prefix)
prefix_re = re.compile("\s*name\s*=\s*.*?\"(?P<prefix>.*?)\|.*?\"")
db = {}


def list_name_prefixes():
    prefixes = set()
    for fn in glob.glob("lesson-files/*"):
        if os.path.isfile(fn):
            s = open(fn, 'r').read()
            for match in prefix_re.finditer(s):
                if match.group('prefix') not in prefixes:
                    print match.group('prefix')
                    prefixes.add(match.group('prefix'))

if options.list_prefixes:
    list_name_prefixes()
    sys.exit()

for fn in glob.glob("lesson-files/*"):
    if os.path.isfile(fn):
        s = open(fn, 'r').read()
        for match in r.finditer(s):
            if match.group('prefix') not in db:
                print "Found prefix:", match.group('prefix')
                db[match.group('prefix')] = [fn]
            else:
                if db[match.group('prefix')][-1] != fn:
                    db[match.group('prefix')].append(fn)
if options.list_strings:
    for s in db.keys():
        print s
        if options.display_filename:
            for fn in db[s]:
                print "\t", fn

print "Skal vi bruke dim eller diminished?"



#!/usr/bin/python

import os
import sys
import re
from subprocess import *
import time


def run(cmd):
    print "run:", cmd
    os.system(cmd)


def get_image_dim(fn):
    output = Popen(["file", fn], stdout=PIPE).communicate()[0]
    r = re.compile("(\d+)\s*x+\s*(\d+)")
    m = r.search(output)
    if m:
        return int(m.groups()[0]), int(m.groups()[1])
    else:
        return None, None


def do_file(fn, width):
    time.sleep(2)
    f, ext = os.path.splitext(fn)
    run("import -frame %s" % fn)
    #x, y = get_image_dim(fn)
    run("pngnq -n 16 -f %s" % fn)
    run("mv %s-nq8.png %s" % (f, fn))
    # if not width:
    #    width = 510
    # if x > width:
    #    run("convert -scale %i %s %s-resized.png" % (width, fn, f))
    #    run("mv %s-resized.png %s" % (f, fn))
    #run("pngquant -nofs 8 %s" % fn)
    #run("mv %s-or8.png %s" % (f, fn))

help = """
Usage: ./tools/screenshot path/to/image.png [width]

Make a screenshot using "import". Run this script, and then
click on the window you want to make a screenshot of.
"""
if len(sys.argv) not in (2, 3):
    print help
    sys.exit()
if sys.argv[1] in ('-h', '--help'):
    print help
    sys.exit()

try:
    width = int(sys.argv[2])
except:
    width = None
do_file(sys.argv[1], width)
print "Remember to use the Simple (Enkelt, nb_NO) theme."



# GNU Solfege - free ear training software
# Copyright (C) 2007, 2008, 2011, 2016 Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class SynthCommon(object):

    def __init__(self):
        self.__test_saved_play_midieventstream = None

    def start_testmode(self):
        assert self.__test_saved_play_midieventstream is None
        self.m_test_stream = None
        self.__test_saved_play_midieventstream = self.play_midieventstream
        self.play_midieventstream = self.testmode_play_midieventstream

    def end_testmode(self):
        assert self.__test_saved_play_midieventstream is not None
        self.play_midieventstream = self.__test_saved_play_midieventstream
        self.__test_saved_play_midieventstream = None
        return self.flush_testdata()

    def flush_testdata(self, details=0):
        if self.m_test_stream:
            return self.m_test_stream.str_repr(details)
        return ""

    def testmode_play_midieventstream(self, midieventstream):
        # twostep operation because we want to sort the events
        # in a specific order
        self.m_test_stream = midieventstream



# vim: set fileencoding=utf-8 :
# GNU Solfege - free ear training software
# Copyright (C) 2007, 2008, 2011, 2016 Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import codecs
import logging
import os
import subprocess
import sys
import textwrap
import time

from gi.repository import GObject
from gi.repository import Gtk

from solfege import buildinfo
from solfege import cfg
from solfege import filesystem
from solfege import gu
from solfege import lessonfile
from solfege import lessonfilegui
from solfege import osutils
from solfege.multipleintervalconfigwidget import IntervalCheckBox
import xml.etree.ElementTree as et

import solfege


def _tr(s):
    """
    Don't try to translate the string if it is empty. This because gettext
    will return the .MO file header as a translation of the empty string.
    """
    return _(s) if s else s


class NameIt(object):

    def __init__(self, name, lilycode):
        self.m_name = name
        self.m_lilycode = lilycode


class Section(list):

    def __init__(self, section_title, line_len):
        self.m_title = section_title
        self.m_line_len = line_len


class BaseSheetWriter(object):
    lilyout = 'lilyout'

    def __init__(self, title):
        self.m_title = title
        self.m_data = []

    def create_outdir(self, directory):
        if not os.path.exists(directory):
            os.mkdir(directory)
        if not os.path.exists(os.path.join(directory, self.lilyout)):
            os.mkdir(os.path.join(directory, self.lilyout))

    def new_section(self, section_title, line_len):
        """
        Return a list where we can append questions.
        """
        v = Section(section_title, line_len)
        self.m_data.append(v)
        return v

    def set_title(self, s):
        self.m_title = s


class HtmlSheetWriter(BaseSheetWriter):

    def _write(self, k, directory, filename, logger):
        assert k in ('question', 'answer')
        f = codecs.open(os.path.join(directory, filename), 'w', 'utf-8')
        print("<html>", file=f)
        print("<!--\n    Generated by GNU Solfege %s\n-->" % buildinfo.VERSION_STRING, file=f)
        print("<link rel='stylesheet' href='style.css' type='text/css'/>", file=f)
        print('<meta http-equiv="Content-Type" content="text/html; charset=utf-8">', file=f)
        print("<body>", file=f)
        print("<h1>%s</h1>" % _tr(self.m_title), file=f)
        for section in self.m_data:
            print("<h2>%s</h2>" % _tr(section.m_title), file=f)
            print("<table>", file=f)
            first_row = True
            for idx, question_dict in enumerate(section):
                if idx % section.m_line_len == 0:
                    if not first_row:
                        print("</tr>", file=f)
                        first_row = False
                    print("<tr>", file=f)
                print("<td class='lilycode'>", file=f)
                # each question is in a table too
                print("<table><tr><td>", file=f)
                if r"\score" in question_dict[k]['music']:
                    print("<lilypond>", file=f)
                else:
                    print("<lilypond fragment relative>", file=f)
                print(question_dict[k]['music'], file=f)
                print("</lilypond>", file=f)
                print("</td></tr><tr><td class='answer'>", file=f)
                print(question_dict[k]['name'], file=f)
                print("</td></tr></table>", file=f)
                print("</td>", file=f)
            print("</table>", file=f)
        print("</body>", file=f)
        f.close()
        args = [cfg.get_string("programs/lilypond-book"),
                "--out", self.lilyout, filename]
        # The lilypond-book.py script is not an executable on MS Windows,
        # so the user have to set the full path to the python script in the
        # preferences window. Then use the same Python interpreter as
        # Solfege itself to run lilypond-book.py
        if sys.platform == 'win32':
            if not os.path.exists(cfg.get_string("programs/lilypond-book")):
                # We to the strerror trick to get the translation for free.
                raise osutils.BinaryForProgramException("lilypond-book",
                    cfg.get_string("programs/lilypond-book"),
                    OSError(2, os.strerror(2)))
            args.insert(0, sys.executable)
        logger.write("\nRunning LilyPond\n", "h1")
        try:
            p = osutils.Popen(args, cwd=directory, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            while 1:
                p.poll()
                # returncode is not None means that the process has finished
                if p.returncode is not None:
                    break
                while 1:
                    s = p.stdout.readline()
                    if not s:
                        break
                    logger.write(s)
                time.sleep(1)
            p.wait()
        except OSError as e:
            raise osutils.BinaryForProgramException("lilypond-book",
                cfg.get_string("programs/lilypond-book"), e)
        f = open(os.path.join(directory, 'README.txt'), 'w')
        print("\n".join(textwrap.wrap("The files answers.html and questions.html in this directory are temporaty files. The HTML files you are looking for are lilyout%(sep)sanswers.html and lilyout%(sep)squestions.html" % {'sep': os.path.sep})), file=f)
        f.close()

    def write_to(self, directory, logger):
        self.create_outdir(directory)
        f = open(os.path.join(directory, self.lilyout, 'style.css'), 'w')
        print("table { border: none }", file=f)
        print(".answer { text-align: center }", file=f)
        f.close()
        self._write('question', directory, 'questions.html', logger)
        self._write('answer', directory, 'answers.html', logger)


class LatexSheetWriter(BaseSheetWriter):

    def write_to(self, directory, logger):
        self.create_outdir(directory)
        self._write('question', directory, 'questions.tex', logger)
        self._write('answer', directory, 'answers.tex', logger)

    def _write(self, k, directory, filename, logger):
        assert k in ('question', 'answer')
        f = open(os.path.join(directory, filename), 'w')
        print(r"\documentclass{article}", file=f)
        print("%%\n%% Created by GNU Solfege %s\n%%" % buildinfo.VERSION_STRING, file=f)
        print(r"\title{%s}" % _tr(self.m_title), file=f)
        print(r"\usepackage[margin=1.0cm]{geometry}", file=f)
        print(r"\begin{document}", file=f)
        print(r"\maketitle", file=f)

        def finish_table(scores, answers):
            for idx, score in enumerate(scores):
                print(score, file=f)
                if idx != len(scores) - 1:
                    print("&", file=f)
            print(r"\\", file=f)
            for idx, answer in enumerate(answers):
                print(answer, file=f)
                if idx != len(scores) - 1:
                    print("&", file=f)
            print(r"\\", file=f)
        for section in self.m_data:
            print(r"\section{%s}" % _tr(section.m_title), file=f)
            first_row = True
            for idx, question_dict in enumerate(section):
                if idx % section.m_line_len == 0:
                    if first_row:
                        print(r"\begin{tabular}{%s}" % ('c' * section.m_line_len), file=f)
                        first_row = False
                    else:
                        finish_table(scores, answers)
                    scores = []
                    answers = []
                if r"\score" in question_dict[k]['music']:
                    scores.append("\n".join((r"\begin{lilypond}",
                                question_dict[k]['music'], r"\end{lilypond}")))
                else:
                    scores.append("\n".join((r"\begin[fragment]{lilypond}",
                                question_dict[k]['music'], r"\end{lilypond}")))
                answers.append(question_dict[k]['name'].replace("#", r"\#"))
            finish_table(scores, answers)
            print(r"\end{tabular}", file=f)
        print(r"\end{document}", file=f)
        f.close()
        # Duplicate of the code in HtmlSheetWriter._write
        args = [cfg.get_string("programs/lilypond-book"),
                "--out", self.lilyout, filename]
        if sys.platform == 'win32':
            if not os.path.exists(cfg.get_string("programs/lilypond-book")):
                # We to the strerror trick to get the translation for free.
                raise osutils.BinaryForProgramException("lilypond-book",
                    cfg.get_string("programs/lilypond-book"),
                    OSError(2, os.strerror(2)))
            args.insert(0, sys.executable)
        logger.write("\nRunning LilyPond\n", "h1")
        try:
            p = osutils.Popen(args, cwd=directory, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            while 1:
                p.poll()
                # returncode is not None means that the process has finished
                if p.returncode is not None:
                    break
                while 1:
                    s = p.stdout.readline()
                    if not s:
                        break
                    logger.write(s)
                time.sleep(1)
            p.wait()
        except OSError as e:
            raise osutils.BinaryForProgramException("lilypond-book",
                cfg.get_string("programs/lilypond-book"), e)
        logger.write("\nRunning LaTeX\n", "h1")
        try:
            p = osutils.Popen(
                (cfg.get_string("programs/latex"), "--interaction=nonstopmode", filename),
                cwd=os.path.join(directory, self.lilyout),
                stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            while 1:
                p.poll()
                # returncode is not None means that the process has finished
                if p.returncode is not None:
                    break
                while 1:
                    while Gtk.events_pending():
                        Gtk.main_iteration()
                    s = p.stdout.readline()
                    if not s:
                        break
                    logger.write(s)
                time.sleep(1)
        except OSError as e:
            raise osutils.BinaryForProgramException("latex",
                cfg.get_string("programs/latex"), e)
        f = open(os.path.join(directory, 'README.txt'), 'w')
        print("\n".join(textwrap.wrap("The files you are looking for are lilyout%(sep)sanswers.dvi and lilyout%(sep)squestions.dvi" % {'sep': os.path.sep})), file=f)
        f.close()


class PractiseSheetDialog(Gtk.Window, gu.EditorDialogBase, lessonfilegui.ExercisesMenuAddIn):
    """
    The definition of which lesson files to create questions from are
    stored in the list PractiseSheetDialog.m_sections. Each item is a
    dict filled with data. See PractiseSheet._add_common for details.

    When an exercise is selected from the menu, on_select_exercise(...)
    is called, and submethods from this will create a set of questions,
    and store the generated questions in m_sections[-1]['questions]
    and the definition (filename, number of questions to generate etc) in
    other items in the dict in PractiseSheetDialog.m_sections.

    When you change the configuration of the exercises, Solfege will only
    generate new random questions if the config changes require new
    questions, for example when question type changes.

    on_create_sheet is called to create the .tex or .html files. Calling
    is several times in a row will generate the same test.

    The file we save from "Ear Training Test Printout" is a combination
    of two things:
        1. The definitons of which exericses to create questions from and
           how many many questions to generate.
        2. The generated questions. So loading the file will let you
           click "Create Sheet" and have the program create the files to
           print out, and you will get the same test as last time you
           did so.

    When you add an exercise, the translated exercise title is used. The
    translated title is stored in the saved file, so it will not change
    language if you restart Solfege in a different locale.
    """
    current_fileformat_version = "2.0"
    STORE_TITLE = 0
    STORE_FILENAME = 1
    ok_music_types = (lessonfile.Chord, lessonfile.Voice, lessonfile.Rvoice)
    ok_modules = ('idbyname', 'harmonicinterval', 'melodicinterval')
    savedir = os.path.join(filesystem.user_data(), "eartrainingtests")

    def __init__(self, filename=None):
        logging.debug("PractiseSheetDialog.__init__")
        Gtk.Window.__init__(self)
        gu.EditorDialogBase.__init__(self, filename)
        self.m_changed = False
        self.set_title(self._get_a_filename())
        self.m_exported_to = None
        self.set_default_size(800, 300)
        # self.vbox contains the toolbar and the vbox with the rest of the
        # window contents
        self.vbox = Gtk.VBox()
        self.add(self.vbox)
        self.setup_toolbar()
        vbox = Gtk.VBox()
        vbox.set_spacing(6)
        self.vbox.pack_start(vbox, True, True, 0)
        vbox.set_border_width(8)
        hbox = Gtk.HBox()
        vbox.pack_start(hbox, False, False, 0)
        hbox.pack_start(Gtk.Label(_("Output format:")), False, False, 0)
        self.g_latex_radio = Gtk.RadioButton.new_with_label(None, "LaTeX")
        hbox.pack_start(self.g_latex_radio, False, False, 0)
        self.g_html_radio = Gtk.RadioButton.new_with_label_from_widget(self.g_latex_radio, "HTML")
        hbox.pack_start(self.g_html_radio, False, False, 0)
        #
        hbox = Gtk.HBox()
        hbox.set_spacing(6)
        vbox.pack_start(hbox, False, False, 0)
        hbox.pack_start(Gtk.Label(_("Title:")), False, False, 0)
        self.g_title = Gtk.Entry()
        hbox.pack_start(self.g_title, False, False, 0)
        #
        self.m_sections = []
        self.g_liststore = Gtk.ListStore(
            GObject.TYPE_STRING,  # lesson-file title
            GObject.TYPE_STRING,  # lessonfile filename, hidden column
        )
        self.g_treeview = Gtk.TreeView(self.g_liststore)
        self.g_treeview.set_size_request(400, 100)
        self.g_treeview.set_headers_visible(False)
        self.g_treeview.connect('cursor-changed', self.on_tv_cursor_changed)
        self.g_treeview.connect('unselect-all', self.on_tv_unselect_all)
        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.add(self.g_treeview)
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        vbox.pack_start(scrolled_window, True, True, 0)
        #
        renderer = Gtk.CellRendererText()

        def mark_invalid(column, cell_renderer, liststore, iter, user_data=None):
            filename = liststore.get(iter, self.STORE_FILENAME)[0]
            if not filename:
                cell_renderer.props.markup = '<span background="red">%s</span>' % liststore.get_value(iter, self.STORE_TITLE)
        column = Gtk.TreeViewColumn(None, renderer, text=self.STORE_TITLE)
        column.set_cell_data_func(renderer, mark_invalid)
        self.g_treeview.append_column(column)
        column = Gtk.TreeViewColumn(None, renderer, text=self.STORE_FILENAME)
        self.g_treeview.append_column(column)
        #
        sizegroup = Gtk.SizeGroup(Gtk.SizeGroupMode.HORIZONTAL)
        self.g_lbox = Gtk.VBox()
        self.g_lbox.set_sensitive(False)
        vbox.pack_start(self.g_lbox, True, True, 0)
        self.g_lesson_title = Gtk.Entry()
        self.g_lbox.pack_start(gu.hig_label_widget(_("Section title:"),
            self.g_lesson_title, sizegroup), False, False, 0)
        self.g_lesson_title_event_handle =\
            self.g_lesson_title.connect('changed', self.on_lesson_title_changed)
        #
        self.g_qtype = Gtk.ComboBoxText()
        # We should not change the order of music types, as the index
        # of the different types are used in SolfegeApp.on_create_sheet
        self.g_qtype.append_text(_("Name the music"))
        self.g_qtype.append_text(_("Write the music, first tone given"))
        self.g_qtype_event_handler = \
            self.g_qtype.connect('changed', self.on_qtype_changed)
        self.g_lbox.pack_start(gu.hig_label_widget(_("Type of question:"),
            self.g_qtype, sizegroup), False, False, 0)
        #
        self.g_intervals = IntervalCheckBox()
        self.g_intervals_box = gu.hig_label_widget(_("Intervals:"),
            self.g_intervals, sizegroup)
        self.g_intervals_event_handler = \
            self.g_intervals.connect('value-changed', self.on_intervals_changed)
        self.g_lbox.pack_start(self.g_intervals_box, False, False, 0)
        #
        self.g_line_len = Gtk.SpinButton(adjustment=Gtk.Adjustment(0, 1, 10, 1, 10))
        self.g_line_len_event_handler = \
            self.g_line_len.connect('value-changed', self.on_spin_changed, 'line_len')
        self.g_lbox.pack_start(gu.hig_label_widget(_("Questions per line:"),
            self.g_line_len, sizegroup), False, False, 0)
        #
        self.g_count = Gtk.SpinButton(adjustment=Gtk.Adjustment(0, 1, 1000, 1, 10))
        self.g_count_event_handler = \
            self.g_count.connect('value-changed', self.on_spin_changed, 'count')
        self.g_lbox.pack_start(gu.hig_label_widget(_("Number of questions:"),
            self.g_count, sizegroup), False, False, 0)
        if filename:
            self.load_file(filename)
        self.add_to_instance_dict()

    def view_lesson(self, idx):
        self.g_qtype.handler_block(self.g_qtype_event_handler)
        self.g_intervals.handler_block(self.g_intervals_event_handler)
        self.g_line_len.handler_block(self.g_line_len_event_handler)
        self.g_count.handler_block(self.g_count_event_handler)
        self.g_lesson_title.handler_block(self.g_lesson_title_event_handle)

        self.g_lbox.set_sensitive(idx is not None)
        if idx is None:
            self.g_lesson_title.set_text("")
            self.g_count.set_value(1)
            self.g_line_len.set_value(1)
        else:
            try:
                module = lessonfile.infocache.get(self.m_sections[idx]['filename'], 'module')
            except lessonfile.infocache.InfoCacheException:
                module = None

            if module == 'harmonicinterval':
                self.g_intervals_box.show()
                self.g_intervals.set_value(self.m_sections[idx]['intervals'])
            else:
                self.g_intervals_box.hide()
            self.g_lesson_title.set_text(_tr(self.m_sections[idx]['title']))
            self.g_count.set_value(self.m_sections[idx]['count'])
            self.g_line_len.set_value(self.m_sections[idx]['line_len'])
            self.g_qtype.set_active(self.m_sections[idx]['qtype'])
        self.g_qtype.handler_unblock(self.g_qtype_event_handler)
        self.g_intervals.handler_unblock(self.g_intervals_event_handler)
        self.g_line_len.handler_unblock(self.g_line_len_event_handler)
        self.g_count.handler_unblock(self.g_count_event_handler)
        self.g_lesson_title.handler_unblock(self.g_lesson_title_event_handle)

    def on_add_lesson_clicked(self, button):
        menu = self.create_learning_tree_menu()
        menu.popup(None, None, None, None, 1, 0)

    def on_remove_lesson_clicked(self, button):
        path, column = self.g_treeview.get_cursor()
        if path is not None:
            iter = self.g_liststore.get_iter(path)
            self.g_liststore.remove(iter)
            del self.m_sections[path[0]]

    def on_randomize(self, widget):
        for section in self.m_sections:
            section['questions'] = []
        self.generate_questions()

    def on_create_sheet(self, widget):
        self.generate_questions()
        if not self.m_exported_to:
            self.m_exported_to = self.select_empty_directory(_("Select where to export the files"))
        if not self.m_exported_to:
            return
        if self.g_html_radio.get_active():
            writer = HtmlSheetWriter(self.g_title.get_text())
        else:
            writer = LatexSheetWriter(self.g_title.get_text())
        for idx, sect in enumerate(self.m_sections):
            new_section = writer.new_section(_(sect['title']), sect['line_len'])
            for question_dict in sect['questions']:
                new_section.append(question_dict)
        ww = gu.LogWindow(self)
        try:
            writer.write_to(self.m_exported_to, ww)
        except osutils.BinaryForProgramException as e:
            solfege.win.display_error_message2(e.msg1, e.msg2)
        ww.run_finished()

    def on_intervals_changed(self, widget, value):
        self.m_changed = True
        idx = self.g_treeview.get_cursor()[0][0]
        self._delete_questions(idx)
        self.m_sections[idx]['intervals'] = value

    def on_lesson_title_changed(self, widget):
        self.m_changed = True
        path = self.g_treeview.get_cursor()[0]
        iter = self.g_liststore.get_iter(path)
        self.m_sections[path[0]]['title'] = widget.get_text()
        self.g_liststore.set(iter, self.STORE_TITLE, widget.get_text())

    def on_spin_changed(self, widget, varname):
        self.m_changed = True
        idx = self.g_treeview.get_cursor()[0][0]
        if varname == 'count':
            if len(self.m_sections[idx]['questions']) > widget.get_value_as_int():
                self.m_sections[idx]['questions'] = \
                    self.m_sections[idx]['questions'][:widget.get_value_as_int()]
        self.m_sections[idx][varname] = widget.get_value_as_int()

    def on_select_exercise(self, menuitem, filename):
        module = lessonfile.infocache.get(filename, 'module')
        if module not in self.ok_modules:
            print("only some modules work:", self.ok_modules)
            return
        p = lessonfile.LessonfileCommon()
        p.parse_file(lessonfile.uri_expand(filename))
        if module == 'idbyname':
            self._add_idbyname_lesson(p, filename)
        elif module == 'harmonicinterval':
            self._add_harmonicinterval_lesson(p, filename)
        elif module == 'melodicinterval':
            self._add_melodicinterval_lesson(p, filename)
        self.g_treeview.set_cursor((len(self.m_sections) - 1,))

    def _add_idbyname_lesson(self, p, filename):
        """
        p is a lessonfile.LessonfileCommon parser that has parsed the file.
        """
        not_ok = len([q.music for q in p.m_questions if not isinstance(q.music, self.ok_music_types)])
        ok = len([q.music for q in p.m_questions if isinstance(q.music, self.ok_music_types)])
        if not_ok > 0:
            if ok > 0:
                do_add = gu.dialog_yesno(_("Not all music types are supported. This file contain %(ok)i supported questions and %(not_ok)i that are not supported. The unsupported questions will be ignored. Add any way?" % locals()), self)
            else:
                gu.dialog_ok(_("Could not add the lesson file. It has no questions with a music object with supported music type."), self)
                do_add = False
        else:
            do_add = True
        if do_add:
            self.m_changed = True
            self._add_common(filename)

    def _add_melodicinterval_lesson(self, p, filename):
        self._add_common(filename)
        self.m_sections[-1]['intervals'] = p.header.ask_for_intervals_0

    def _add_harmonicinterval_lesson(self, p, filename):
        self._add_common(filename)
        self.m_sections[-1]['intervals'] = p.header.intervals

    def _add_common(self, filename):
        self.m_changed = True
        self.m_sections.append({
            'filename': filename,
            'title': lessonfile.infocache.get(filename, 'title'),
            'count': 6,
            'qtype': 0,
            'line_len': 3,
            'questions': [],
        })
        self.g_liststore.append((_(self.m_sections[-1]['title']), filename))

    def generate_questions(self):
        """
        Delete any extra questions and then add new random questions
        if some sections have too few questions generated.
        """
        for section in self.m_sections:
            # Remove trailing questions if we have too many questions
            del self.m_sections[section['count']:]
            # Then add questions if we have too few
            count = section['count'] - len(section['questions'])
            if count:
                for question_dict in solfege.app.sheet_gen_questions(
                        count, section):
                    section['questions'].append(question_dict)

    def _delete_questions(self, idx):
        """
        Delete the generated questions for exercise idx in self.m_sections.
        """
        self.m_sections[idx]['questions'] = []

    def on_show_help(self, widget):
        solfege.app.handle_href("ear-training-test-printout-editor.html")

    def on_tv_cursor_changed(self, treeview, *v):
        if self.g_treeview.get_cursor()[0] is None:
            return
        self.view_lesson(self.g_treeview.get_cursor()[0][0])

    def on_tv_unselect_all(self, treeview, *v):
        self.view_lesson(None)

    def on_qtype_changed(self, widget):
        self.m_changed = True
        idx = self.g_treeview.get_cursor()[0][0]
        self._delete_questions(idx)
        self.m_sections[idx]['qtype'] = widget.get_active()

    def load_file(self, filename):
        tree = et.ElementTree()
        tree.parse(filename)
        if tree.getroot().get("fileformat_version") != "2.0":
            e = Exception("PractiseSheet")
            e.msg1 = _("Cannot read old file format")
            e.msg2 = _("To convert the file to the new file format, you must open and save it in an older version of Solfege. Versions from 3.16.0 to 3.20.4 should do the job.")
            raise e
        # section.find('text').text will be none if it was not set
        # when saved, and we need a string.
        if tree.find("title").text:
            self.g_title.set_text(tree.find("title").text)
        else:
            self.g_title.set_text("")
        if tree.find('output_format').text == 'latex':
            self.g_latex_radio.set_active(True)
        else:
            self.g_latex_radio.set_active(False)
        self.g_liststore.clear()
        self.m_sections = []
        for section in tree.findall("section"):
            d = {}
            lessonfilename = section.find('filename').text
            # It seems that the elementtree parser will return str if
            # there are no non-ascii chars in the filename. So lets
            # make unicode of it if it is str
            if isinstance(lessonfilename, str):
                lessonfilename = str(lessonfilename)
            d['filename'] = lessonfilename
            # section.find('text').text will be none if it was not set
            # when saved, and d['title'] need to be a string
            if section.find('title').text:
                d['title'] = section.find('title').text
            else:
                d['title'] = ""
            if section.find('intervals') is not None:
                d['intervals'] = eval(section.find('intervals').text, {}, {})
            else:
                if lessonfile.infocache.get(lessonfilename, 'module') == 'harmonicinterval':
                    d['intervals'] = []
                    gu.dialog_ok("FIXME: «%s» was saved with a buggy version of solfege, so you must set the intervals by selecting the file in the dialog and clicking the intervals to be asked. Sorry!" % lessonfilename, self)
            d['count'] = int(section.find('count').text)
            d['line_len'] = int(section.find('line_len').text)
            d['qtype'] = int(section.find('qtype').text)
            d['questions'] = []
            for question in section.findall("question"):
                q = {'question': {}, 'answer': {}}
                q['question']['music'] = question.find("students").find("music").text
                q['question']['name'] = question.find("students").find("name").text
                q['answer']['music'] = question.find("teachers").find("music").text
                q['answer']['name'] = question.find("teachers").find("name").text
                d['questions'].append(q)
            self.m_sections.append(d)
            try:
                # Check that the filename is valid
                lessonfile.infocache.get(lessonfilename, 'title')
                self.g_liststore.append((d['title'], lessonfilename))
            except lessonfile.infocache.InfoCacheException as e:
                self.g_liststore.append((_("«%s» not found") % str(e), None))
            self.g_treeview.set_cursor((0,))
        self.m_filename = filename

    def save(self):
        assert self.m_filename
        if self.g_latex_radio.get_active():
            format = "latex"
        else:
            format = "html"
        doc = et.Element("sheet", fileformat_version=self.current_fileformat_version,
                         creator="GNU Solfege",
                         app_version=buildinfo.VERSION_STRING)
        doc.append(et.Comment(""))
        et.SubElement(doc, "title").text = self.g_title.get_text()
        et.SubElement(doc, "output_format").text = format
        for sect in self.m_sections:
            s = et.Element("section")
            et.SubElement(s, "title").text = sect['title']
            et.SubElement(s, "filename").text = sect['filename']
            # only the harmonic and melodic intervals have this variable:
            if 'intervals' in sect:
                et.SubElement(s, 'intervals').text = "%s" % sect['intervals']
            # We do save 'count' because this variable say how many questions
            # we want to generate, and len(questions) will just say how many
            # are generated now.
            et.SubElement(s, "count").text = "%i" % sect['count']
            et.SubElement(s, "line_len").text = "%i" % sect['line_len']
            et.SubElement(s, "qtype").text = "%i" % sect['qtype']
            for qdict in sect['questions']:
                q = et.SubElement(s, "question")
                t = et.SubElement(q, "teachers")
                et.SubElement(t, "name").text = qdict['answer']['name']
                et.SubElement(t, "music").text = qdict['answer']['music']
                t = et.SubElement(q, "students")
                et.SubElement(t, "name").text = qdict['question']['name']
                et.SubElement(t, "music").text = qdict['question']['music']
            doc.append(s)
        tree = et.ElementTree(doc)
        f = open(self.m_filename, 'w')
        print('<?xml version="1.0" encoding="utf-8"?>', file=f)
        tree.write(f, encoding="utf-8")
        f.close()
        self.m_changed = False

    def setup_toolbar(self):
        self.g_toolbar = Gtk.Toolbar()
        self.g_actiongroup.add_actions([
         ('Add', Gtk.STOCK_ADD, None, None, None, self.on_add_lesson_clicked),
         ('Remove', Gtk.STOCK_REMOVE, None, None, None, self.on_remove_lesson_clicked),
         ('Create', Gtk.STOCK_EXECUTE, _("Create Sheet"), None, None, self.on_create_sheet),
         ('Randomize', None, _("Randomize"), None, None, self.on_randomize),
        ])
        self.g_ui_manager.insert_action_group(self.g_actiongroup, 0)
        uixml = """
        <ui>
         <toolbar name='ExportToolbar'>
          <toolitem action='Add'/>
          <toolitem action='Remove'/>
          <toolitem action='New'/>
          <toolitem action='Open'/>
          <toolitem action='Save'/>
          <toolitem action='SaveAs'/>
          <toolitem action='Create'/>
          <toolitem action='Randomize'/>
          <toolitem action='Close'/>
          <toolitem action='Help'/>
         </toolbar>
         <accelerator action='Close'/>
         <accelerator action='New'/>
         <accelerator action='Open'/>
         <accelerator action='Save'/>
        </ui>
        """
        self.g_ui_manager.add_ui_from_string(uixml)
        self.vbox.pack_start(self.g_ui_manager.get_widget("/ExportToolbar"),
                             False, False, 0)
        self.g_ui_manager.get_widget("/ExportToolbar").set_style(Gtk.ToolbarStyle.BOTH)



#!/usr/bin/python3
# coding: iso-8859-1

import re
import sys
import optparse
import unittest
import glob
import os

ok_different_mapping_key = (
  "notenameformat|%(notename)s%(oct)s",
)


class Entry(object):
    # translated status: UNTRANS, FUZZY, TRANS
    # obsolete: True / False

    def __init__(self, s):
        s = s.strip("\n")
        msgid_re = re.compile('(msgid|#~ msgid) "(?P<id>.*?)"\s*$')
        msgstr_re = re.compile('(msgstr|#\~ msgstr) "(?P<str>.*?)"\s*$')
        data = {'comment': [], 'msgid': [], 'msgstr': []}
        mode = None
        self.m_status = 'translated'
        self.m_obsolete = False
        for line in s.split("\n"):
            msgid_m = msgid_re.match(line)
            msgstr_m = msgstr_re.match(line)
            if line.startswith("#:") or line.startswith("#.") or line.startswith("#, python-format") or line.startswith("#, no-python-format") or line.startswith("# ") or line.startswith("#|") or line == '#':
                assert mode in ('comment', None)
                mode = 'comment'
                data[mode].append(line)
            elif line.startswith("#, fuzzy"):
                assert mode in ('comment', None)
                mode = 'comment'
                self.m_status = 'fuzzy'
                data[mode].append(line)
            elif msgid_m:
                assert mode in ('comment', None)
                mode = 'msgid'
                if line.startswith("#~"):
                    self.m_obsolete = True
                if msgid_m.group('id'):
                    data[mode].append(msgid_m.group('id'))
            elif msgstr_m:
                assert mode == 'msgid'
                mode = 'msgstr'
                if msgstr_m.group('str'):
                    data[mode].append(msgstr_m.group('str'))
            else:
                # We are in a mlineline msgid or msgstr
                assert mode in ('msgid', 'msgstr'), mode
                data[mode].append(line.strip('"'))
        self.m_msgid = data['msgid']
        self.m_msgstr = data['msgstr']
        self.m_comment = data['comment']
        if not "".join(self.m_msgstr):  # self.m_msgstr == ['']:
            self.m_status = 'untranslated'
        if not self.m_msgid:
            self.m_status = 'heading'
        for idx in range(len(self.m_msgid)):
            if self.m_msgid[idx] == '':
                self.m_msgid[idx] = '\n'
        self.m_msgid_string = "".join(self.m_msgid)
        for idx in range(len(self.m_msgstr)):
            if self.m_msgstr[idx] == '':
                self.m_msgstr[idx] = '\n'
        self.m_msgstr_string = "".join(self.m_msgstr)
        #    translated = 'heading'
        # if not "\n".join(retval['msgstr']):
        #    translated = 'untranslated'
        # return translated, obsolete, retval

    def is_header(self):
        return not self.m_msgid

    def get_status(self):
        """
        'translated'
        'untranslated'
        'fuzzy'
        'obsolete'
        """
        if self.m_obsolete:
            return 'obsolete'
        return self.m_status

    def get_msgid(self):
        return self.m_msgid_string
        return " ".join(self.m_msgid)

    def get_msgstr(self):
        return self.m_msgstr_string
        return " ".join(self.m_msgstr)


class TestEntry(unittest.TestCase):

    def test_escaped_qoute(self):
        s = """
#: solfege/dictation.py:197
#, python-format
msgid "The '%s'."
msgstr "Il file \"%s\"."

"""
        e = Entry(s)
        self.assertEqual(e.get_msgstr(), "Il file \"%s\".")

def sniff_po_encoding(filename):
    r = re.compile(b".*Content-Type: text/plain; charset=(.*?)\\\\n\"")
    with open(filename, "br") as f:
        for line in f.readlines():
            m = r.match(line)
            if m:
                return m.groups()[0].decode("ascii")


def string_to_entry_list(s):
    """
    This function assumes that the file is well formed. There should
    be one empty line between each message."
    """
    v = []
    for x in s.split("\n\n"):
        if not x:
            print("NOTHIN")
            continue
        e = Entry(x)
        v.append(e)
    return v


def parse_string(s):
    """
    This function will return a dict with some statistics about the entries.
    """
    v = string_to_entry_list(s)
    dicts = {'translated': 0, 'fuzzy': 0, 'obsolete': 0, 'untranslated': 0, 'heading': 0}
    for e in v:
        dicts[e.get_status()] += 1
    del dicts['heading']
    return dicts


def parse_file(filename):
    with open(filename, 'r') as f:
        return parse_string(f.read())


def find_string_formatting(s):
    """
    Make a list with all string formatting operators in the string except
    named operators, like %(name)s
    """
    s = re.sub("%\s*%", "", s)
    #r = re.compile("\%(\(\w+\))?(\.?\d+)?\w")
    r = re.compile("\%(\.?\d+)?[^\s\(]")
    start = 0
    m = r.search(s[start:])
    ret = []
    while m:
        ret.append(m.group())
        start = start + m.end()
        m = r.search(s[start:])
    return ret


def find_string_formatting_mapping_keys(s):
    s = re.sub("%\s*%", "", s)
    r = re.compile("\%(\((?P<name>\w*)\))?(?P<notname>(\.?\d+)?[^\s\(])")
    start = 0
    m = r.search(s[start:])
    ret = []
    while m:
        ret.append((m.group('name'), m.group('notname')))
        start = start + m.end()
        m = r.search(s[start:])
    return ret


class TestFindStringFormatting(unittest.TestCase):

    def test_simplest(self):
        for n in ("%s", "%f", "%.2d", "%03d"):
            self.assertEqual(find_string_formatting(n), [n])

    def test_complex(self):
        self.assertEqual(find_string_formatting("abc %.1f%%\n def %.1f%%"), ['%.1f', '%.1f'])

    def test_common_po_bugs(self):
        self.assertNotEqual(find_string_formatting("%.1f%.\n"), ['%.1f'])

    def test_double_percent(self):
        self.assertEqual(find_string_formatting("%s %% %f"), ['%s', '%f'])
        self.assertEqual(find_string_formatting("%s % % %f"), ['%s', '%f'])
        self.assertEqual(find_string_formatting("%s %\t\t% %f"), ['%s', '%f'])

    def test_ignore(self):
        self.assertEqual(find_string_formatting("%(name)s"), [])

    def test_xxx(self):
        msgid = """
Test completed!
Your score was %.1f%%.
The test requirement was %.1f%%.
"""
        msgstr = """
Úspěšný test!
Vaše skóre bylo %.1f%%.
Požadované skóre bylo%.1f% %.
"""
        self.assertEqual(find_string_formatting(msgid),
                         find_string_formatting(msgstr))

    def test_mapping(self):
        v1 = find_string_formatting_mapping_keys("%(name)s %(bla).1f X")
        v2 = find_string_formatting_mapping_keys("%(bla).1f X X %(name)s ")
        v1.sort()
        v2.sort()
        self.assertEqual(v1, v2)


def check_file(filename):
    enc = sniff_po_encoding(filename)
    with open(filename, 'r', encoding=enc) as f:
        v = string_to_entry_list(f.read())
    file_ok = True
    for entry in v:
        if entry.is_header():
            continue
        a = find_string_formatting(entry.get_msgid())
        b = find_string_formatting(entry.get_msgstr())
        if (a != b) and entry.get_msgstr() and \
                (entry.get_status() not in ('obsolete', 'fuzzy')):
            file_ok = False
            print("%s:" % filename)
            print('msgid: "%s"' % entry.get_msgid())
            print('msgstr: "%s"' % entry.get_msgstr)()
            print("%s != %s" % (a, b))
            print("Entry status:", entry.get_status())
            print()
        # Check mapping key
        v1 = find_string_formatting_mapping_keys(entry.get_msgid())
        v2 = find_string_formatting_mapping_keys(entry.get_msgstr())
        all_1 = v1 and bool([x for x in v1 if x[0] is not None])
        all_2 = v2 and bool([x for x in v2 if x[0] is not None])
        v1x = list(v1)
        v2x = list(v2)
        v1x.sort()
        v2x.sort()
        format_ok = (v1x == v2x and all_1 and all_2)
        if (v1 != v2) and entry.get_msgstr() \
                and (entry.get_status() not in ('obsolete', 'fuzzy')) \
                and not [s for s in entry.m_comment if 'no-python-format' in s] \
                and (entry.get_msgid() not in ok_different_mapping_key) \
                and (not format_ok):
            file_ok = False
            print("Mapping key not equal in the file", filename)
            print("\tmsgid: \"" + entry.get_msgid() + "\"")
            print("\tmsgstr: \"" + entry.get_msgstr() + "\"")
        # Check for notename| strings
        if entry.get_msgid().startswith("notename|"):
            if "|" in entry.get_msgstr():
                file_ok = False
                print("%s:" % filename)
                print("'|' in translated notename entry:")
                print('msgid: "%s"' % entry.get_msgid())
    return file_ok


if len(sys.argv) > 1 and sys.argv[1] == '-t':
    del sys.argv[1]
    unittest.main()
    sys.exit()
if len(sys.argv) > 1 and sys.argv[1] == '-c':
    with open('po/tr.po', 'r') as f:
        s = f.read()
    for x in s.split("\n\n"):
        print("\n", x)
    sys.exit()


retval = 0
for filename in glob.glob(os.path.join("po", "*.po")):
    if not check_file(filename):
        retval = -1
sys.exit(retval)
